#!/bin/sh
#
# A non-global zone can be in one of the following six states:
#
# Incomplete
#  During an install or uninstall operation, zoneadm sets the state of the target zone to incomplete.
#  Upon successful completion of the operation, the state is set to the correct state.
#
# Configured
#  The zone's configuration is complete and committed to stable storage.
#  Note here - those elements of the zone's application environment that must be specified after initial boot are not yet present.
#
# Installed
#  The zone's configuration is instantiated on the system.
#  The zoneadm command is used to verify that the configuration can be successfully used on the designated Solaris system.
#  Packages are installed under the zone's root path. In this state, the zone has no associated virtual platform.
#
# Ready
#  The virtual platform for the zone is established.
#  The kernel creates the zsched process, network interfaces are set up and made available to the zone, file systems are mounted, and devices are configured.
#  A unique zone ID is# assigned by the system. At this stage, no processes associated with the zone have been started.
#
# Running
#  User processes associated with the zone application environment are running. The zone enters the running state as soon as the first user process associated with the application environment (init) is created.
#
# Shutting down and Down
#  These states are transitional states that are visible while the zone is being halted. However, a zone that is unable to shut down for any reason will stop in one of these states.
#
#

. /opt/HAC/bin/rsf.sh
#
# args: <start|stop> [no_of_attempts]
#
state=$1                # starting or stopping?
attempts=${2:-'1'}      # no of attempts at this action
service=$3
zonename=$4
tenant=$4

servername="`uname -n`"
script="`basename $0`"

#
# Get the value  of the pool mount  point so we  can check
# the mountpoint is contained in the zone root mount point.
pool_mountpoint=`zfs get -H -o value mountpoint ${service}`

zoneroot=`nawk '/zonepath/{
    gsub(/\"/, " ");
    for (i = 1; i < NF; i++)
        if ($i ~ /zonepath/)
            print $(i+1)
}' /${service}/.zones/${zonename}/${zonename}.xml`

#
# NEX-21661:
# This is a defensive check to ensure  the pool  mountpoint is contained
# in the zone root - we have seen cases where  tenant information stored
# in the pools .zones directory contained tenants info from another pool
# and this the scrips shut down tenants running on that  different pool.
if expr "${zoneroot}" : "${pool_mountpoint}/.*" >/dev/null 2>&1
then
    dated_echo "Found zoneroot for zone ${zonename}: ${zoneroot}"
else
    dated_echo "Wrong pool for ${zoneroot} - should be ${pool_mountpoint} - aborting ${state} operation on this tenant"
    exit ${RSF_OK}
fi

#
# you may or may not want this test
#
if [ ${attempts} -gt 3 ]; then
    # fail after three goes
    dated_echo "${service}: Too many retries on ${state}, aborting"
    exit ${RSF_ABORT}       # abort code
fi

#
# NEX-19337 - add the -p flag to preserve publishers.
ZONEATTACH_PRESERVE_FLAG="-p"

#
# Search through the global zones list of BE's and search for the
# currently active one.
HOST_UUID=""
globalGetActiveBootEnvUUID()
{
    HOST_UUID=""
    for line in `beadm list -H`; do
        active=`echo $line|awk -F";" '{print $3}'`
        if [ "NR" == "${active}" ] || [ "N" == "${active}" ] ; then
            HOST_UUID=`echo ${line} | awk -F";" '{print $2}'`
            break
        fi
    done
}

sanitiseInstalledZone()
{
    tenant_path=`zonecfg -z ${zonename} info zonepath | awk '{print $2}'`
    tenant_mntpnt="${tenant_path}/root"
    tenant_ds=`mount|grep "^${tenant_path}" | awk '{print $3}'`
    tenant_root="${tenant_ds}"

    globalGetActiveBootEnvUUID

    if [ -z "${HOST_UUID}" ] ; then
        dated_echo "${service}: sanitiseInstalledZone - unable to locale global zone UUID"
        return
    else
        zlist=`zfs list -rH ${tenant_ds} 2>/dev/null | awk '{print $1}'`
        for zelement in ${zlist}; do
            is_active=`zfs get -H org.opensolaris.libbe:active ${zelement} | awk '{ print $3}'`
            is_mpt_set=`zfs get -H mountpoint ${zelement} | awk '{ print $2 }'`
            if [ ${is_active} == "on" ] && [ ${is_mpt_set} != "legacy" ]; then
                dated_echo "${service}: sanitiseInstalledZone - found active dataset with non legacy mountpoint - fixing"
                zfs set zoned=off ${zelement}
                zfs set mountpoint=legacy ${zelement}
                zfs set org.opensolaris.libbe:parentbe=${HOST_UUID} ${zelement}
                zfs set zoned=on ${zelement}
                return
            fi
        done
    fi
}

findactive()
{
    tenant_path=`zonecfg -z ${zonename} info zonepath | awk '{print $2}'`
    if [ -z "${tenant_path}" ] ; then
        dated_echo "*** Unable to read tenant path from zonecfg"
        return
    fi
    tenant_mntpnt="${tenant_path}/root"
    dated_echo "Tenant mount point is ${tenant_mntpnt}"

    if grep -q ${tenant_mntpnt} /etc/mnttab; then
        dated_echo "Tenant root is already mounted - no need to force mount"
        return
    fi

    tenant_ds=`mount | grep "^${tenant_path} " | awk '{print $3}'`
    tenant_root="${tenant_ds}/ROOT"
    dated_echo "Tenant root: $tenant_root"

    lvers=`pkg info entire | grep FMRI | awk '{print $2}'`
    dated_echo "entire version: ${lvers}"

    zlist=`zfs list -rH ${tenant_root} | awk '{print $1}'`
    dated_echo "List of datasets: `echo ${zlist} | tr '\n' ' '`"

    for x in ${zlist}; do
        dated_echo "Determining if dataset ${x} is active"
        is_active=`zfs get -H org.opensolaris.libbe:active $x | awk '{print $3}'`

        if [ "${is_active}" == "on" ]; then
            dated_echo "Dataset ${x} is active - mounting"
            mount -F zfs ${x} ${tenant_mntpnt}

            dated_echo "Mount complete for ${x} - finding tanant entire version"
            zvers=`pkg -R ${tenant_mntpnt} info entire | grep FMRI | awk '{print $2}'`

            if [ "${zvers}" == "${lvers}" ]; then
                # Found the correct active tenant BE leave it mounted.
                dated_echo "${x} has correct entire version: ${zvers}"
                return
            else
                dated_echo "${x} has incorrect entire version: ${zvers} (not ${lvers}) - unmounting"
                umount ${tenant_mntpnt}
            fi
        else
            dated_echo "Dataset ${x} is not active"
        fi
    done

    dated_echo "*** Failed to find active tenant BE"
    exit 0
}

#
# Import zone configuration from SUNWdetached.xml file.
zoneimport()
{
    # overwrite existing zone confguration
    dated_echo "Importing zone configuration for ${zonename}"
    zonecfg -z ${zonename} create -Fa ${zoneroot}
    dated_echo "Zone import complete for ${zonename}"
}

zoneattach()
{
    # check if zone exists in zone repository
    if [ `zoneadm list -ic | grep -cw ${zonename}` -eq 1 ]; then
        # capture current zone state
        zonestate=`zoneadm -z ${zonename} list -p | awk -F: '{print $3}'`
    else
        dated_echo "Zone ${zonename} does not exist in ${servername} zone repository"
        dated_echo "Zone ${zonename} is unknown on ${servername}, preparing to import zone configuration ..."
        # explicitly set new zone state to unconfigured
        zonestate="unconfigured"
    fi

    case "${zonestate}" in

        # handle both configured and unconfigured, i.e. new, zones
        configured|unconfigured)
            # check if SUNWdetached.xml configuration file exists
            if [ -f ${zoneroot}/SUNWdetached.xml ]; then
                dated_echo "SUNWdetached.xml file exists, importing latest zone configuration ..."
                zoneimport
            else
                if [ ${zonestate} = "unconfigured" ]; then
                    dated_echo "ALERT: SUNWdetached.xml file does not exist so cannot import zone configuration, aborting ..."
                    return
                fi
            fi

            dated_echo "Attaching zone ${zonename} to ${servername} global zone"

            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="attach" state="start"

            #
            # NEX-17433: Fix tenant/zone root not available under certain circumstances due
            # to other system panicing.
            findactive
            # NEX-17433: end
            #
            dated_echo "Running zoneadm attach for ${zonename}"
            if zoneadm -z ${zonename} attach ${ZONEATTACH_PRESERVE_FLAG}; then
                dated_echo "Zone ${zonename} succesfully attached"
                /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="attach" state="complete"
            else
                dated_echo "ALERT: Zone ${zonename} could not be attached, aborting ..."
                /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="attach" state="failed"
            fi
            ;;

        'installed')
            if [ -f /opt/HAC/RSF-1/etc/LIVEUPGRADE ]; then
                dated_echo "ALERT: Live Upgrade mode detected"
                dated_echo "Zone ${zonename} is already attached to ${servername} global zone ..."
            else
                /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="attach" state="start"
                #
                # If a zone is in installed state, this could be a throwback
                # to a crashed instance; check the zone for sanity and update
                # the zone state top off, legacy mountpoint and the global BE
                # UUID.
                sanitiseInstalledZone

                /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="attach" state="complete"
                dated_echo "WARNING: Zone ${zonename} is installed on ${servername}, zone is already attached ..."
                # check if SUNWdetached.xml configuration file exists
                if [ -f /export/zones/${zonename}/SUNWdetached.xml ]; then
                    dated_echo "WARNING: SUNWdetached.xml file exists, zone is already attached so cannot import latest zone configuration ..."
                fi
            fi
            ;;

        *)
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="attach" state="start"
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="attach" state="failed"
            dated_echo "ALERT: Zone ${zonename} is ${zonestate} on ${servername}, aborting ..."
            ;;

    esac
}

zonedetach()
{
    # capture current zone state
    zonestate=`zoneadm -z ${zonename} list -p | awk -F: '{print $3}'`
    case "${zonestate}" in

        'installed')
            if [ -f /opt/HAC/RSF-1/etc/LIVEUPGRADE ]; then
                dated_echo "ALERT: Live Upgrade mode detected"
                dated_echo "Leaving zone ${zonename} attached to ${servername} global zone ..."
            else
                /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="detach" state="start"
                dated_echo "Detaching zone ${zonename} from ${servername} global zone"
                if zoneadm -z ${zonename} detach; then
                    /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="detach" state="complete"
                    dated_echo "Zone ${zonename} successfully detached"
                else
                    /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="detach" state="failed"
                    dated_echo "ALERT: Zone ${zonename} could not be detached, aborting ..."
                fi
            fi
            ;;

        'configured')
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="detach" state="start"
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="detach" state="complete"
            dated_echo "WARNING: Zone ${zonename} is configured on ${servername}, zone is already detached ..."
            ;;

        *)
            dated_echo "ALERT: Zone ${zonename} is ${zonestate} on ${servername}, aborting ..."
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="detach" state="start"
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="detach" state=${zonestate}
            ;;

    esac
}

zoneboot()
{
    zonestate=`zoneadm -z ${zonename} list -p | awk -F: '{print $3}'`
    #
    # NEX-18144 - pull out the zone HA attribute from the zone XML file.
    # If set to false then don't boot the zone just leave it in an atached state.
    ZONEHAATTR=`zonecfg -z ${zonename} info attr name=ha|sed -n 's/^.*value: *//p'`
    if [ -z "${ZONEHAATTR}" ] ; then
        ZONEHAATTR="true"
    fi

    if [ "${ZONEHAATTR}" = "false" ] ; then
        dated_echo "Zone ${zonename} has boot attribute set to false - not booting"
        return
    fi

    dated_echo "Zone ${zonename} is ${zonestate} on ${servername}, booting zone ..."
    dated_echo "Transitioning zone ${zonename} into running state, please wait ..."

    /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="boot" state="start"

    if zoneadm -z ${zonename} boot; then
        dated_echo "zoneadm -z ${zonename} boot complete"
    else
        dated_echo "WARNING: Failed to verify zone XML configuration file, aborting ..."
        /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="boot" state="failed"
        return
    fi

    zonestate=`zoneadm -z ${zonename} list -p | awk -F: '{print $3}'`
    case "${zonestate}" in

        'running')
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="boot" state="complete"
            dated_echo "Zone ${zonename} is now running on ${servername}, zone successfully booted"
            #
            # Make a copy of the zone config to zone detached file.
            ;;

        'shutting_down')
            dated_echo "ALERT: Zone ${zonename} is shutting down on ${servername}, aborting ..."
            dated_echo "Zone ${zonename} is shutting down, please wait ..."
            dated_echo "Elapsed time (secs):\c"; timer=0

            while zoneadm -z ${zonename} list -p | grep ${zonestate} > /dev/null
            do
                sleep 1
                timer=`expr ${timer} + 1`
                echo " ${timer}\c"

            done
            echo "\r"

            zonestate=`zoneadm -z ${zonename} list -p | awk -F: '{print $3}'`
            dated_echo "Zone ${zonename} is now ${zonestate} on ${servername}"
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="boot" state=${zonestate}
            ;;

        'down')
            dated_echo "ALERT: Zone ${zonename} is down on ${servername}, aborting ..."
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="boot" state="failed"
            ;;

        *)
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="boot" state=${zonestate}
            dated_echo "ALERT: Zone ${zonename} is ${zonestate} on ${servername}, aborting ..."
            ;;

    esac
}

zonehalt()
{
    zonestate=`zoneadm -z ${zonename} list -p | awk -F: '{print $3}'`
    /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="shutdown" state="start"
    case "${zonestate}" in
        'running')
            dated_echo "Zone ${zonename} is running on ${servername}, shutting down zone ..."
            #
            # NEX-17840 changed init 5 to halt
            zoneadm -z ${zonename} halt
            dated_echo "Zone ${zonename} shutdown started, please wait ..."
            dated_echo "Elapsed time (secs):\c"; timer=0

            while zoneadm list | grep -w ${zonename} > /dev/null
            do
                sleep 1
                timer=`expr ${timer} + 1`
                echo " ${timer}\c"
            done
            echo "\r"

            zonestate=`zoneadm -z ${zonename} list -p | awk -F: '{print $3}'`
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="shutdown" state="complete"
            dated_echo "Zone ${zonename} is now ${zonestate} on ${servername}, zone successfully shutdown"
            ;;

        installed|configured)
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="shutdown" state="complete"
            dated_echo "WARNING: Zone ${zonename} is ${zonestate} on ${servername}, zone is already shutdown"
            ;;

        *)
            /opt/HAC/RSF-1/bin/event_notifier "LOG_INFO" "RSF_TENANT" zone="${zonename}" operation="shutdown" state="aborted"
            dated_echo "Zone ${zonename} is ${zonestate} on ${servername}, aborting ..."
            ;;

    esac
}

#
# decide action based on first argument
#
case "${state}" in

    'start')
        zoneattach
        zoneboot
        ;;

    'stop')
        zonehalt
        zonedetach
        ;;

    *)
        echo "Usage: $0 <start|stop>"
        exit ${RSF_WARN}        # warning code
        ;;
esac

exit ${RSF_OK}  # OK code (default)
