/* =============================================================================
**  Filename:   librsfapi.h
**
**  Copyright:  High-Availability.com Ltd (c)
**
**  Website:    http://www.high-availability.com/
** =============================================================================
*/
#ifndef LIBRSFAPI_H_
#define LIBRSFAPI_H_

/*
** API Version info
** ----------------
*/
#define API_VERSION "4.2.1"

/*
** Default networking options
** --------------------------
**
** DEFAULT_IP is the address used to communicate with rsfmon if no other address
** is provided. RSF_PORT is the port used to communicate with rsfmon if no other
** port is provided. DEFAULT_ATTEMPTS is the maximum number of times a service
** command is attempted before abandoning the operation.
*/
#define DEFAULT_IP "127.0.0.1"
#define RSF_PORT "1195"
#define DEFAULT_ATTEMPTS 5
#define MOVE_ATTEMPTS 20

/*
** Timeout for networking communications during node test in seconds.
*/
#define TEST_NODE_TIMEOUT 2

/*
** Seed for blowfish
** -----------------
*/
#define PASSWD_SEED 123456L

/*
** T-flag support
** --------------
**
** Service packet information is returned by rsfmon using either 'S' or T' lines
** depending on the version. Information on the 'T' lines is more comprehensive
** than the 'S' lines. Some early versions of rsfmon, however, return service
** information packets on 'S' lines. Using 'T' flags is reccommended if using a
** recent rsfmon version.
*/
#define RSF_KNOWS_T_FLAG

/*
** Halt mode support
** -----------------
**
** Support for the halt command can be disabled if the API is built with a
** version of rsfmon that does not support it.
*/
#ifdef DISABLE_HALT_MODE
// Disable 'halt' command
#ifdef RSF_KNOWS_HALT
#undef RSF_KNOWS_HALT
#endif
#else
// Enable 'halt' command
#ifndef RSF_KNOWS_HALT
#define RSF_KNOWS_HALT
#endif
#endif

#define HALT_ERR "Support for the 'halt' command is disabled for this build"

/* =============================================================================
** Debug File Preprocessing
** ------------------------
**
** RSFAPI_DEBUG_FILE is the location to use for the the API log file.
** 
** RSFAPI_DEBUG_ON_FILE is the location of the API debugging trigger file. If
** this file exists, medium level API debugging will be enabled. Low or high
** level debugging can be enabled by if the trigger file has "-low" or "-high"
** appended to the file name respectively. In the event that multiple trigger
** files exist, "-high" is preferred, followed by "-low" then by the standard
** RSFAPI_DEBUG_FILE filename.
** 
** RSFLOG is an internal convienence function to for debug file logging.
** RSFERR is an intenral convienence function to log API errors.
*/
#define RSFAPI_DEBUG_FILE "/opt/HAC/RSF-1/log/rsfapi.log"
#define RSFAPI_DEBUG_ON_FILE "/tmp/rsfapi-debug-on"

#ifndef RSFLOG
#define RSFLOG(X) rsf1::rsfDebug.write ( std::ostringstream().flush() << X )
#endif

#ifndef RSFERR
#define RSFERR(X,Y) rsf1::rsfDebug.error ( X, std::ostringstream().flush() << Y )
#endif

#ifndef RSFEXCEPT
#define RSFEXCEPT(X,Y) rsf1::rsfDebug.except ( X, Y )
#endif

#ifndef PRINT_CLI_ERRORS
#define PRINT_CLI_ERRORS rsf1::rsfDebug.cliMode
#endif

/* 
** Convienence check to determine if current debug level permits the preceeding
** action.
** 
** DEBUG_LOW is triggered when the debug level is 1 or higher.
** DEBUG_MED is triggered when the debug level is 2 or higher.
** DEBUG_HIGH is triggered when the debug level is 3.
** 
** DEBUG_INFO
**
*/
#define DEBUG_LOW   if (rsf1::rsfDebug.getDebugLevel() >= 1 )
#define DEBUG_MED   if (rsf1::rsfDebug.getDebugLevel() >= 2 )
#define DEBUG_HIGH  if (rsf1::rsfDebug.getDebugLevel() == 3 )
#define DEBUG_INFO "[" << __FUNCTION__ << ":" << __LINE__ << "]: "

/*
** Returns the current API error value set.
*/
#define RSFERRNO rsf1::rsfDebug.getError()

/* =============================================================================
** Colour printing support
** -----------------------
**
** SET_COL sets the output colour according ot the value of ansi_colour_codes
** provided.
*/
#ifndef SETCOL
#define SETCOL(X) "\033[" << X << "m"
#endif

/* =============================================================================
** Include files
** -------------
*/
#include <algorithm>
#include <cstddef>
#include <dirent.h>
#include <errno.h>
#include <exception>
#include <fcntl.h>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <netdb.h>
#include <ostream>
#include <signal.h>
#include <sstream>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>
#include <vector>

#include <arpa/inet.h>

#include <ifaddrs.h>
#include <netinet/in.h>

/*
** FreeBSD specific include files
*/
#if defined(__FreeBSD__)
#include <sys/types.h>
#include <netinet/in.h>
#endif

#include <pwd.h>

/*
** Include the required headers from rsflib
*/
#include "pwd_file.h"
#include "librsf.h"
#include "rsfcon.h"


/* =============================================================================
** Namespace declaration
** ---------------------
** 
** To prevent poluuting the std namespace, all API classes are decalred within
** the rsf1 namespace.
*/
namespace rsf1
{
  using namespace std;

/* =============================================================================
** RSF Error code enumerator
** -------------------------
** 
** All valid response codes from the API are defined here.
*/

 /**
  * @brief      API response codes enumerator.
  */
  enum rsf_error_enum
  {
    RSF_OK = 0,

    // Authentication errors
    //
    RSF_AUTH_DIGEST,        // No digest provided in authentication
    RSF_AUTH_GENDIGEST,     // Failed to compute message digest
    RSF_AUTH_INVALID,       // Invalid authentication vector detected

    // CLI errors
    //
    RSF_CLI_DBGLVL,         // Invalid debug level provided to CLI
    RSF_CLI_INVALID_AF,     // Invalid address family preference provided to CLI
    RSF_CLI_INVALID_ARG,    // Invalid argument provided to CLI command
    RSF_CLI_KILL_FAILED,    // Failed to send kill signal to rsfmon process
    RSF_CLI_LOCALONLY,      // Requested CLI command doesn't support local only mode
    RSF_CLI_QUIT,           // CLI command requested termination
    RSF_CLI_NO_ARG,         // No required argument passed to CLI command
    RSF_CLI_PARSE_TOUT,     // Failed to parse the timeout provided to CLI
    RSF_CLI_PIDERROR,       // Unable to open the PID file
    RSF_CLI_TOO_FEW_ARGS,   // Not enough arguments passed to CLI command
    RSF_CLI_TOO_MANY_ARGS,  // Too many arguments passed to CLI command
    RSF_CLI_UNKNOWN_CMD,    // Unrecognised command passed to CLI
    RSF_CLI_UNKNOWN_FLAG,   // Unrecognised flag passed to CLI
    RSF_CLI_VERSINFO,       // Failed to get version info for rsf-1

    // rsfcomms() error
    //
    RSF_COMMS_NORESPONSE,   // No response was received from rsf
    RSF_COMMS_RSFERR,       // RSF-1 returned an error for the command request
    RSF_COMMS_ALREADY,      // Already doing what you are asking"
    RSF_COMMS_NOTRUN,       // Not running: Service to be killed/restarted"
    RSF_COMMS_FAIL,         // Something failed (bad password etc.)"
    RSF_COMMS_PASSFAIL,     // Password Match Failed"
    RSF_COMMS_BADSERVICE,   // Unknown Service: Can't set action"
    RSF_COMMS_NOT_BROKEN,   // Not Broken: Can't Repair"
    RSF_COMMS_BUSY,         // Another request in progress, try poll time later
    RSF_COMMS_ARGS,         // Bad or missing args in request"
    RSF_COMMS_LOCKED,       // Failed: You don't own this nodes keylock"
    RSF_COMMS_DATA,         // Data Follows"
    RSF_COMMS_SYSDOWN,      // System shutting down"
    RSF_COMMS_SALT,         // Failed to receive salt from host
    RSF_COMMS_BADRESPONSE,  // Got a bad response from rsf

    // Empty errors
    //
    RSF_EMPTY_COMMAND,      // Empty command passed to function
    RSF_EMPTY_SUBCMD,       // Nothing passed as sub-command
    RSF_NO_CLUSTER,         // No cluster detected
    RSF_NO_DATA,            // No data detected in response
    RSF_NO_NODES,           // No clustered nodes detected
    RSF_NO_RESPONSE,        // No reply received from host

    // Exception errors
    //
    RSF_EXCEPT_ALLINFO,     // Exception occurred - getAllLetterInfoHandler
    RSF_EXCEPT_GETQUADS,    // Exception occurred - getQuads
    RSF_EXCEPT_QUADINT,     // Exception occurred - getQuadElementInt
    RSF_EXCEPT_QUADSTR,     // Exception occurred - getQuadElement
    RSF_EXCEPT_SVCCMD,      // Exception occurred - Service command
    RSF_EXCEPT_SVCINT,      // Exception occurred - Service_int
    RSF_EXCEPT_SVCMVE,      // Exception occurred - Service_move

    // Fork errors
    //
    RSF_FORK_FAIL,          // Failed to fork a child process

    // File errors
    //
    RSF_FILE_REMOVE,        // Failed to remove file
    RSF_FILE_WRITE,         // Failed to write to file

    // Feature disabled errors
    //
    RSF_HALT_DISABLED,      // Halt support disabled

    // Invalid errors
    //
    RSF_INVALID_ADDR,       // Failed to validate address
    RSF_INVALID_CLUSTER,    // Broken cluster detected

    // Move command errors
    //
    RSF_MVE_MANFAIL,        // Failed to set services to manual during service move
    RSF_MVE_STOPFAIL,       // Failed to stop service during service move
    RSF_MVE_STARTFAIL,      // Failed to start service during service move
    RSF_MVE_INVALIDATTEMPTS,// Invalid attempt value provided for service move

    // Obtain errors
    //
    RSF_DEMO_MODE_ERR,      // Failed to obtain demo mode status
    RSF_POLLTIME_ERROR,     // Failed to obtain polltime from cluster

    // Parse errors
    //
    RSF_PARSE_STRTOINT,     // Failed to convert string to integer

    // Range errors
    //
    RSF_RANGE_INITIMEOUT,   // Initial timeout value out of range
    RSF_RANGE_RUNTIMEOUT,   // Runtime timeout value out of range

    // RSF cache errors
    //
    RSF_RNC_CORRUPTENTRY,   // Corrupt cache entry detected
    RSF_RNC_INFOVALUE,      // Failed to obtain value from the cache
    RSF_RNC_INVALIDFLAG,    // Invalid flag provided to internal enum
    RSF_RNC_LETTER_ID,      // Invalid letter ID detected
    RSF_RNC_OUTOFRANGE,     // Provided out of range ID for value
    RSF_RNC_UNKNOWNNODE,    // Failed to find node entry in cache
    RSF_RNC_CLEARCACHE,     // Failed to clear cache entries
    RSF_RNC_CLEARHOST,      // Failed to clear cached localhost name

    // Socket errors
    //
    RSF_SOCK_CONNECT,       // Failed to connect to socket
    RSF_SOCK_CREATE,        // Failed to create socket
    RSF_SOCK_GETOPTS,       // Failed to obtain socket options
    RSF_SOCK_NTOP,          // Call to inet_ntop() failed
    RSF_SOCK_RECV,          // Failed to receive data from socket
    RSF_SOCK_RESOLVE_HOST,  // Failed to resolve host
    RSF_SOCK_SELECT,        // Call to select() failed
    RSF_SOCK_SEND,          // Failed to send data to socket
    RSF_SOCK_SETOPT,        // Failed to set socket option
    RSF_SOCK_TIMEOUT,       // Socket timeout reached

    // Service errors
    //
    RSF_SVC_ALREADYRUNNING, // Service is already running on target node
    RSF_SVC_ATTEMPT_EXPIRE, // Reached maximum number of service command attempts
    RSF_SVC_INVALID,        // Invalid service ID provided
    RSF_SVC_NOINSTANCES,    // No service instances found
    RSF_SVC_UNCONFNODE,     // Node is not configured to run service
    RSF_SVC_UNKNOWN,        // Failed to locate service

    // Unknown property errors
    //
    RSF_UNKNOWN_AFINET,     // Unknown address family
    RSF_UNKNOWN_EXTCMD,     // Unknown extended service command provided
    RSF_UNKNOWN_NODE,       // Node is unknown to the cluster
    RSF_UNKNOWN_SUBCMD,     // Unknown secondary service command provided
    RSF_UNKNOWN_SVCCMD,     // Unknown service command provided

    RSF_ERR_COUNT
  };

/* =============================================================================
** RSF Cache Enumerators
** ---------------------
** 
** The API has multiple enumerators defined to make cache allocations and
** access operations easier.
*/

 /**
  * @brief      Emunerator for the valid RSF response 'letters' that the API
  *             accepts in the internal cache.
  */
  enum rsf_info_letters
  {
    rsfInfo_MISC = 0,
    rsfInfo_B,
    rsfInfo_C,
    rsfInfo_F,
    rsfInfo_G,
    rsfInfo_H,
    rsfInfo_I,
    rsfInfo_N,
    rsfInfo_P,
    rsfInfo_R,
    rsfInfo_S,
    rsfInfo_V,
    rsfInfo_W,
    rsfInfo_COUNT
  };

 /**
  * @brief      Miscellaneous cache value identifiers.
  */
  enum rsfInfo_MISC_values
  {
    MISC_called_node = 0,
    MISC_COUNT
  };

 /**
  * @brief      [B]ackup cache value identifiers.
  */
  enum rsfInfo_B_values
  {
    b_service_number = 1,
    b_can_run_here,
    b_LISTALL_servers,
    b_SIZE_OK
  };

 /**
  * @brief      [C]luster cache value identifiers.
  */ 
  enum rsfInfo_C_values
  {
    c_cluster_name = 1,
    c_node_count,
    c_service_count,
    c_config_file_crc,
    c_serial_number,
    c_config_file_mtime,
    c_config_file_size,
    c_poll_time,
    c_startup_time,
    c_machine_bootstrap,
    c_machine_ident,
    c_machine_desc,
    c_SIZE_OK
  };

 /**
  * @brief      [F]lag cache value identifiers.
  */ 
  enum rsfInfo_F_values
  {
    f_cFlag = 1,
    f_rFlag,
    f_sFlag,
#ifdef RSF_KNOWS_HALT
    f_hFlag,
#endif
    f_SIZE_OK
  };

 /**
  * @brief      Confi[G]ured heartbeat cache value identifiers.
  */ 
  enum rsfInfo_G_values
  {
    g_heartbeat_id = 1,
    g_heartbeat_type,
    g_source_machine,
    g_dest_machine,
    g_dest_address,
    g_extra_param1, /* Net:=  Destination IP    / Disc:=  Initial Backoff     */
    g_extra_param2, /* Net:=  via Hostname:Port / Disc:=  Backoff Multiplier  */
    g_extra_param3, /* Net:=  via IP Address    / Disc:=  Backoff Upper Limit */
    g_SIZE_OK
  };

 /**
  * @brief      [H]eartbeat cache value identifiers.
  */  
  enum rsfInfo_H_values
  {
    h_heartbeat_id = 1,
    h_heartbeat_status,
    h_sequence_num,
    h_update_time,
    h_SIZE_OK
  };

 /**
  * @brief      [I]nstances cache value identifiers.
  */ 
  enum rsfInfo_I_values
  {
    i_service_number = 1,
    i_state,
    i_priority,
    i_unblocked,
    i_initimeout,
    i_runtimeout,
    i_controlunblocked,
    i_service_name,
    i_restricted,
    i_last_change_time,
    i_state_change_count,
    i_SIZE_OK
  };

 /**
  * @brief      [N]odes cache value identifiers.
  */ 
  enum rsfInfo_N_values
  {
    n_machine_name = 1,
    n_ip_address,
    n_inet_family,
    n_SIZE_OK
  };

 /**
  * @brief      [P] Extra network interface cache value identifiers.
  */ 
  enum rsfInfo_P_values
  {
    p_machine_name = 1,
    p_service_number,
    p_device_name,
    p_LIST_ADDR_TYPE,
    p_LIST_ADDRESS,
    p_LIST_NETMASK,
    p_SIZE_OK
  };

 /**
  * @brief      [R]elease cache value identifiers.
  */ 
  enum rsfInfo_R_values
  {
    r_release_version = 1,
    r_release_date,
    r_release_string,
    r_patch_level,
    r_pkg_version,
    r_SIZE_OK
  };

 /**
  * @brief      [T] [S]ervices cache value identifiers.
  */ 
  enum rsfInfo_S_values
  {
    s_service_number = 1,
    s_service_name,
    s_floatname,
#ifdef RSF_KNOWS_T_FLAG
    s_netmask,
#endif
    s_description,
    s_QUAD_SERVER,
    s_QUAD_IPDEVICE,
    s_QUAD_INITIMEOUT,
    s_QUAD_RUNTIMEOUT,
    s_SIZE_OK
  };

 /**
  * @brief      RSF-1 [V]ersion cache value identifiers.
  */ 
  enum rsfInfo_V_values
  {
    v_client_protocol = 1,
    v_rsf_release,
    v_restricted_node,
    v_license_expire_time,
    v_system_type,
    v_license_type,
    v_machine_id,
    v_license_check_result,
    v_SIZE_OK
  };

 /**
  * @brief      [W]hichNode cache value identifiers.
  */ 
  enum rsfInfo_W_values
  {
    w_machine_name = 1,
    w_hostname,
    w_host_ID,  // A patch is applied if the W line is missing the host ID.
    w_ip_address,
    w_inet_family,
    w_SIZE_OK
  };

/* =============================================================================
** RSF Discovery Options
** ---------------------
** 
** Options for API cluster discovery are listed below. Cluster discovery uses
** broadcast packets to detect active clusters/nodes on a given subnet.
*/

 /**
  * @brief      Enumerated list of possible cluster discovery modes
  */
  enum rsf_discovery_modes
  {
    rsf_discover_all = 0,
    rsf_discover_bootstrap,
    rsf_discover_cluster
  };

/* =============================================================================
** RSF Cluster Actions
** -------------------
** 
** The list of supported cluster actions is below. These are the actions that
** can be performed for a service under cluster control.
*/

 /**
  * @brief      Emumerated list of cluster service actions.
  */
  enum rsf_cluster_actions
  {
    rsf_action_bounce = 0,  // B
    rsf_action_timeout,     // D
    rsf_action_restart,     // E
    rsf_action_pass,        // F
    rsf_action_stop,        // G
    rsf_action_manual,      // H
    rsf_action_auto,        // I
    rsf_action_halt,        // JA
    rsf_action_bstop,       // K
    rsf_action_shutdown,    // L
    rsf_action_repair,      // U
    rsf_action_block,       // W
    rsf_action_unblock,     // X
    rsf_action_log,         // Y
    rsf_action_recovered,   // [
    rsf_action_quickstart,  // @AQ

    rsf_action_COUNT
  };

/* =============================================================================
** RSF Heartbeat Types
** -------------------
** 
** The list of supported heartbeat types is below. These are used by private
** heartbeat functions to determine which heartbeats should be considered during
** a type search.
*/

/**
 * @enum          rsf_heartbeat_type
 * 
 * @brief         Enumerator of valid heartbeat types.
 */
enum rsf_heartbeat_type
{
  HBT_NET = 0,  /*  Any network heartbeat         */
  HBT_NET4,     /*  IPv4 network heartbeat        */
  HBT_NET6,     /*  IPv6 network heartbeat        */
  HBT_SERIAL,   /*  Serial heartbeat              */
  HBT_DISC,     /*  Disc heartbeat                */
  HBT_ANY       /*  Any heartbeat type            */
};

/* =============================================================================
** Coloured output enumerator
** --------------------------
*/

 /**
  * @brief      This enumerator contains the codes used to format output to the
  *             console with colour. Codes beginning with FG_ changes the text
  *             forground colour, codes beginning with BG_ changes the text
  *             background colour.
  */
  enum ansi_colour_codes
  { 
    FG_BLACK    = 30,
    FG_RED      = 31,
    FG_GREEN    = 32,
    FG_YELLOW   = 33,
    FG_BLUE     = 34,
    FG_MAGENTA  = 35,
    FG_CYAN     = 36,
    FG_WHITE    = 37,
    FG_DEFAULT  = 39,
    BG_BLACK    = 40,
    BG_RED      = 41,
    BG_GREEN    = 42,
    BG_YELLOW   = 43,
    BG_BLUE     = 44,
    BG_MAGENTA  = 45,
    BG_CYAN     = 46,
    BG_WHITE    = 47,
    BG_DEFAULT  = 49
  };

/* =============================================================================
** API Classes
** -----------
** 
** This API is divided up into classes. Each class groups together functions
** that relate to a common dataset/functionality.
*/

 /**
  * @brief      Class for API authentication commands.
  */
  class Auth
  {
   /*
    * Need to be friends with RSFNet so rsfcomms so authenticated commands can
    * be performed.
    */
    friend class RSFNet;

  private:
    Auth( const Auth& other );      // non construction-copyable
    Auth& operator=( const Auth& ); // non copyable

   /*
    * Digest functions
    * ----------------
    */
    void   asciiToDigest( string Ascii, UWORD_32bits *Digest );
    string getPwdDigest ( string pass ); 
    string msgHexDigest ( unsigned * msgDigest );
    string msgDigest    ( unsigned * msgDigest );

   /*
    * Admin authentication
    * --------------------
    */    
    string getAdminDigest ();
    string getAdminHash   ( string user = RSF_ADMIN_USER );
    
   /*
    * Authentication generation
    * -------------------------
    */ 
    string generateAuthline(
      vector<string> authvec,
      string salt,
      string request );

   /*
    * Interactive functions
    * ---------------------
    */     
    string getUserPass ( string prompt );

   /*
    * Binary functions
    * ----------------
    */
    string getBin   ( string str  );
    string intToBin ( unsigned long n );

  public:
    Auth();  // Auth Constructor
    vector<string> genObject(
      string user = "", string pass = "", string digest = "" );

  };

 /**
  * @brief      Class for obtaining information for the cluster backup servers.
  */
  class Backup
  {
  public:
    Backup();     // Backup Constructor
    ~Backup();    // Backup Destructor

   /*
    * Data store size
    * ---------------
    */
    unsigned getCount( string node = DEFAULT_IP );

   /*
    * Data store access
    * -----------------
    */
    int            getServiceNumber( int sp = 0, string node = DEFAULT_IP );
    int            getState        ( int sp = 0, string node = DEFAULT_IP );
    vector<string> getServers      ( int sp = 0, string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string getAllBackupInfo( string node = DEFAULT_IP, bool nl = true );
  };

 /**
  * @brief      Class for obtaining information for an RSF cluster.
  */
  class Cluster
  {
  public:
    Cluster();     // Cluster Constructor
    ~Cluster();    // Cluster Destructor

   /*
    * Store access functions
    * ----------------------
    */
    string  getClusterName      ( string node = DEFAULT_IP );
    int     getNodeCount        ( string node = DEFAULT_IP );
    int     getServiceCount     ( string node = DEFAULT_IP );
    string  getConfigCRC        ( string node = DEFAULT_IP );
    string  getSerialNumber     ( string node = DEFAULT_IP );
    int     getConfigMakeTime   ( string node = DEFAULT_IP );
    int     getConfigSize       ( string node = DEFAULT_IP );
    int     getPollTime         ( string node = DEFAULT_IP );
    int     getStartupTime      ( string node = DEFAULT_IP );
    int     getMachineBootstrap ( string node = DEFAULT_IP );
    string  getClusterID        ( string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string  getAllClusterInfo   ( string node = DEFAULT_IP, bool nl = true );
  };


 /**
  * @brief      Class for logging any errors that may occur within the API.
  */
  class Debug
  {
  private:
    Debug( const Debug& other );      // non construction-copyable
    Debug& operator=( const Debug& ); // non copyable

    static bool     firstWrite;
    static unsigned rsfError;
    static int      debuglvl;

   /*
    * Output stream for the log file
    * ------------------------------
    */ 
    std::ofstream logfile;

   /*
    * Initial check function
    * ----------------------
    */ 
    void _initialChecks();

  public:
    Debug();     // Debug Constructor

    static bool cliMode;
    static bool useColour;

   /*
    * API error functions
    * -------------------
    */ 
    int getError();
    int setError( rsf_error_enum status );

   /*
    * Debugging level functions
    * -------------------------
    */ 
    int getDebugLevel();
    int setDebugLevel( unsigned lvl );

   /*
    * Error message string
    * --------------------
    */ 
    string rsfErrorString( rsf_error_enum err );
    string getTraceStr( char letter,  string fn );

   /*
    * Debug streams
    * -------------
    */
    void write( std::ostream& s );
    void error( std::string trace, std::ostream& s );
    void except( std::string trace, std::exception& e );
  };


 /**
  * @brief      Class for obtaining extra network interface configuration information.
  */
  class ExtraNetInfo
  {
    vector<string> getListValues(
      rsfInfo_P_values value, unsigned idx, string node );

    string getListElement(
      rsfInfo_P_values key, unsigned idx, unsigned sp, string node );

  public:
    ExtraNetInfo();     // ExtraNetInfo Constructor
    ~ExtraNetInfo();    // ExtraNetInfo Destructor

   /*
    * Data store size
    * ---------------
    */
    unsigned getCount( string node = DEFAULT_IP );
   /*
    * Data store access functions
    * ---------------------------
    */
    string getMachineName  ( int sp = 0, string node = DEFAULT_IP );
    int    getServiceNumber( int sp = 0, string node = DEFAULT_IP );
    string getDeviceName   ( int sp = 0, string node = DEFAULT_IP );

   /*
    * Repeated list functions
    * -----------------------
    */
    unsigned getListSize( unsigned sp, string node = DEFAULT_IP );

    string getIPType  ( int sp = 0, unsigned idx = 0, string node = DEFAULT_IP );
    string getHostname( int sp = 0, unsigned idx = 0, string node = DEFAULT_IP );
    string getNetmask ( int sp = 0, unsigned idx = 0, string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string getAllHBConfigInfo( string node = DEFAULT_IP, bool nl = true );
  };

 /**
  * @brief      Class for obtaining information for the RSF flags currently in effect.
  */
  class Flags
  {
  private:
    int getFlagStatus( char flag, string node, string _fn );

  public:
    Flags();     // Flags Constructor
    ~Flags();    // Flags Destructor

   /*
    * Flag status functions
    * ---------------------
    */
    int getCRCFlag        ( string node = DEFAULT_IP );
    int getRestrictedFlag ( string node = DEFAULT_IP );
    int getSysShutdownFlag( string node = DEFAULT_IP );
    int getHaltFlag       ( string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */
    string getAllFlagInfo( string node = DEFAULT_IP, bool nl = true );
  };

 /**
  * @brief      Class for obtaining configured heartbeat information.
  * 
  * Note that a configured heartbeat will have back off information if (and
  * only if) its type is DISC. The backoff variables will be the default 
  * the configured heartbeat is not of type DISC.
  * which is -1. 
  */
  class HBConfig
  {
  private:
    int getExtraParamInt(
      rsfInfo_G_values value,
      int              sp,
      string           node,
      string           _fn );

    string getNetChar(
      rsfInfo_G_values value,
      int              sp,
      string           node,
      string           _fn );

    vector<unsigned> getHBList(
      string _fn,
      rsf_heartbeat_type type,
      string node );

  public:
    HBConfig();     // HBConfig Constructor
    ~HBConfig();    // HBConfig Destructor

   /*
    * Store counter function
    * ----------------------
    */ 
    unsigned getCount( string node = DEFAULT_IP );

   /*
    * Enumerator functions
    * --------------------
    */
    vector<string> getHeartbeatTypes( rsf_heartbeat_type type );

   /*
    * Counter functions
    * -----------------
    */
    unsigned getNetCount   ( string node = DEFAULT_IP );   
    unsigned getNetCount4  ( string node = DEFAULT_IP );    
    unsigned getNetCount6  ( string node = DEFAULT_IP );    
    unsigned getSerialCount( string node = DEFAULT_IP );
    unsigned getDiscCount  ( string node = DEFAULT_IP );

   /*
    * List functions
    * --------------
    */ 
    vector<unsigned> getNetList    ( string node = DEFAULT_IP  );
    vector<unsigned> getNetList4   ( string node = DEFAULT_IP  );
    vector<unsigned> getNetList6   ( string node = DEFAULT_IP  );
    vector<unsigned> getSerialList ( string node = DEFAULT_IP  );
    vector<unsigned> getDiscList   ( string node = DEFAULT_IP  );

   /*
    * Store access functions
    * ----------------------
    */
    int    getID                 ( int sp = 0, string node = DEFAULT_IP );
    string getType               ( int sp = 0, string node = DEFAULT_IP );
    string getSourceMachine      ( int sp = 0, string node = DEFAULT_IP );
    string getDestinationMachine ( int sp = 0, string node = DEFAULT_IP );
    string getDestinationAddress ( int sp = 0, string node = DEFAULT_IP );
    int    getIniBackOff         ( int sp = 0, string node = DEFAULT_IP );
    int    getBackOffMultiplier  ( int sp = 0, string node = DEFAULT_IP );
    int    getBackOffUpperLimit  ( int sp = 0, string node = DEFAULT_IP );
    string getNetDestinationIP   ( int sp = 0, string node = DEFAULT_IP );
    string getNetViaHostPortCombo( int sp = 0, string node = DEFAULT_IP );
    string getNetViaIP           ( int sp = 0, string node = DEFAULT_IP );
    
   /*
    * All info function
    * -----------------
    */ 
    string getAllHBConfigInfo( string node = DEFAULT_IP, bool nl = true );
  };

 /**
  * @brief      Class for obtaining the status of cluster heartbeats.
  */
  class Heartbeat
  {
  private:
    vector<unsigned> getHBList(
      string _fn, rsf_heartbeat_type type, string node = DEFAULT_IP );

  public:
    Heartbeat();  // Heartbeat Constructor
    ~Heartbeat(); // Heartbeat Destructor

   /*
    * Store counter function
    * ----------------------
    */
    unsigned getCount( string node = DEFAULT_IP );

   /*
    * Counter functions
    * -----------------
    */ 
    unsigned getNetCount   ( string node = DEFAULT_IP );
    unsigned getNetCount4  ( string node = DEFAULT_IP );
    unsigned getNetCount6  ( string node = DEFAULT_IP );
    unsigned getSerialCount( string node = DEFAULT_IP );
    unsigned getDiscCount  ( string node = DEFAULT_IP );

   /*
    * List functions
    * --------------
    */ 
    vector<unsigned> getNetList   ( string node = DEFAULT_IP );
    vector<unsigned> getNetList4  ( string node = DEFAULT_IP );
    vector<unsigned> getNetList6  ( string node = DEFAULT_IP );
    vector<unsigned> getSerialList( string node = DEFAULT_IP );
    vector<unsigned> getDiscList  ( string node = DEFAULT_IP );

   /*
    * Read functions
    * --------------
    */ 
    int     getID            ( int sp = 0, string node = DEFAULT_IP );
    string  getDiskStatus    ( int sp = 0, string node = DEFAULT_IP );
    int     getSequenceNumber( int sp = 0, string node = DEFAULT_IP );
    int     getUpdateTime    ( int sp = 0, string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string  getAllHBeatInfo   (string node = DEFAULT_IP, bool nl = true);
  };

 /**
  * @brief      Class for obtaining information for an instance of a service in
  *             the cluster.
  * 
  * Creates an instance object from a service number, service mode, initial
  * timeout, run timeout, service name, service state, unblocked boolean value,
  * control unblocked boolean value, restriction, last change time, state change
  * count, and additional information.
  */
  class Instances
  {
   /**
    * @brief      Enumerated list of valid cluster restriction types.
    */
    enum restricted_state
    {
      unrestricted = 0,   /**< unrestricted                          */
      config,             /**< configuration restricted              */
      licence,            /**< licence restricted                    */
      config_and_licence, /**< configuration and licence restricted  */
      not_read            /**< corrupt data detected - data not read */
    };

   /*
    * Read functions
    * --------------
    */ 
    string strRestricted( int value );

  public:
    Instances();     // Instances Constructor
    ~Instances();    // Instances Destructor

   /*
    * Store ID functions
    * ------------------
    */
    unsigned getIdForServiceNumber( int serviceNumber, string host );

   /*
    * Store counter function
    * ----------------------
    */
    unsigned getCount( string node = DEFAULT_IP );

   /*
    * Store access functions
    * ----------------------
    */ 
    int    getServiceNumber   ( int sp = 0, string node = DEFAULT_IP );
    string getState           ( int sp = 0, string node = DEFAULT_IP );
    int    getStateNum        ( int sp = 0, string node = DEFAULT_IP );
    int    getPriority        ( int sp = 0, string node = DEFAULT_IP );
    int    getUnblocked       ( int sp = 0, string node = DEFAULT_IP );
    int    getIniTimeout      ( int sp = 0, string node = DEFAULT_IP );
    int    getRunTimeout      ( int sp = 0, string node = DEFAULT_IP );
    int    getControlUnblocked( int sp = 0, string node = DEFAULT_IP );
    string getServiceName     ( int sp = 0, string node = DEFAULT_IP );
    string getRestricted      ( int sp = 0, string node = DEFAULT_IP );
    int    getLastChangeTime  ( int sp = 0, string node = DEFAULT_IP );
    int    getStateChangeCount( int sp = 0, string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string  getAllReleaseInfo  ( string node = DEFAULT_IP, bool nl = true );

   /*
    * Locate service
    * --------------
    */ 
    vector<string> locateService( string serviceName, string host = DEFAULT_IP );
    vector<string> locateServiceNumber( int serviceNumber, string host = DEFAULT_IP );

   /*
    * Enquiry functions
    * -----------------
    */
    int isServiceBlocked( string service, string node );
  };

 /**
  * @brief      Class for obtaining information for a cluster node.
  * 
  * Creates a node object from a node name, node IP and any additional
  * information.
  */
  class Node
  {
  public:
    Node();     // Node Constructor
    ~Node();    // Node Destructor

   /*
    * Store counter function
    * ----------------------
    */
    unsigned getCount( string node = DEFAULT_IP );

   /*
    * Store access functions
    * ----------------------
    */ 
    string  getMachineName  ( int sp = 0, string node = DEFAULT_IP );
    string  getIP           ( int sp = 0, string node = DEFAULT_IP );
    string  getAddressFamily( int sp = 0, string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string  getAllNodeInfo( string node = DEFAULT_IP, bool nl = true );

   /*
    * List functions
    * --------------
    */ 
    vector<string> getNodesList (
      string host        = DEFAULT_IP,
      bool   showIP      = false,
      bool   hideLocal   = false,
      bool   dynamicMode = false );
  };

 /**
  * @brief      Class for obtaining release version information from RSF.
  * 
  * Creates a version object from the client protocol, rsf release,
  * manual only restriction flag, licence expire date, system type,
  * machine ID and the signature type.
  */
  class Release
  {
  public:
    Release();     // Release Constructor
    ~Release();    // Release Destructor

   /*
    * Read functions
    * --------------
    */ 
    string getVersion       ( string = DEFAULT_IP );
    string getRelDate       ( string = DEFAULT_IP );
    string getRelString     ( string = DEFAULT_IP );
    string getPatchLevel    ( string = DEFAULT_IP );
    string getPackageVersion( string = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string getAllReleaseInfo( string node = DEFAULT_IP, bool nl = true );
  };

 /**
  * @brief      Class to provide actions for the cluster and its services
  */
  class RSFActions
  {
    bool validateServer ( string host );

    string lockCmd (
      string         cmd,
      vector<string> authvec,
      string         address,
      string         data    = "",
      string         rsfdata = "" );

    string fileCmd (
      string         cmd,
      vector<string> authvec,
      string         address,
      string         data    = "",
      string         rsfdata = "" );

    vector<string> processAuth(
      vector<string> authvec, string username, string password );

    int performAction (
      string              service,
      string              address,
      unsigned            attempts,
      vector<string>      authvec,
      string              username   = "",
      string              password   = "",
      rsf_cluster_actions action     = rsf_action_COUNT,
      int                 svcNum     = -1,
      string              cmdSuffix  = ""  );

  public:
    RSFActions();         // RSFActions Constructor
    ~RSFActions();        // RSFActions Destructor

    int validateInstance ( string srv, string addr );

   /*
    * Service Actions
    * ---------------
    */ 
    int serviceManual(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceAuto(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceStop(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = ""
      );

    int serviceBlockStop(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceBlock(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceUnblock(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceBounce(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int servicePass(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceRestart(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceRepair(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");


    int serviceTimeout(
      string         service,
      string         address    = "",
      string         initimeout = "",
      string         runtimeout = "",
      unsigned       attempts   = DEFAULT_ATTEMPTS,
      vector<string> authvec    = vector<string>(),
      string         username   = "",
      string         password   = "");

    int serviceQuickstart(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceMove(
      string         service_name,
      string         dest_ip,
      vector<string> authvec   = vector<string>(),
      string         username  = "",
      string         password  = "",
      unsigned       timeout   = 10,
      int            quickmove = 0,
      int            sticky    = 0);

    int serviceQuickmove(
      string         service_name,
      string         dest_ip,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "",
      unsigned       timeout  = 10,
      int            sticky   = 0);

    int serviceShutdown(
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceRecovered(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceHalt(
      string         service,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    int serviceLog(
      string         msg_type,
      string         log_level,
      string         message,
      string         address,
      unsigned       attempts = DEFAULT_ATTEMPTS,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

   /*
    * Locking Commands
    * ----------------
    */
    string lockNode(
      string          address,
      string          seconds,
      vector<string>  authvec  = vector<string>(),
      string          username = "",
      string          password = "" );

    string unlockNode(
      string         address,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

   /*
    * Config file actions
    * -------------------
    */ 
    string configNew(
      string         address,
      string         data     = "",
      string         crc      = "",
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    string configRollback(
      string         address,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    string configCommit(
      string         address,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

   /*
    * Cluster restart
    * ---------------
    */ 
    string coldRestart(
      string         address,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

    string hotRestart(
      string         address,
      vector<string> authvec  = vector<string>(),
      string         username = "",
      string         password = "");

   /*
    * File actions
    * ------------
    */ 
    string fileDirectory(
      string address, string directory, string username, string password );

    string fileGet(
      string address, string filename, string username, string password );

    string fileStats(
      string address, string filename, string username, string password );

    friend class CLI;
  };

 /**
  * @brief      Class for RSF API networking
  */
  class RSFNet
  {
  private: 
    RSFNet( const RSFNet& other );      // non construction-copyable
    RSFNet& operator=( const RSFNet& ); // non copyable

   /*
    * Network timeout variables
    * -------------------------
    */ 
    int timeoutSec;
    int timeoutUsec;

   /*
    * Connection timeout function
    * ---------------------------
    */
    int connectTimeout  ( int sockfd, const struct sockaddr *addr, socklen_t addrlen,
      int to_sec, int to_usec);

  public:
    RSFNet();     // RSFNet Constructor

   /*
    * IP Address family
    * -----------------
    */ 
    int pref_AF;

   /*
    */
    void setTimeout ( float timeout );

   /*
    * Validation/test functions
    * -------------------------
    */
    int testNode ( string address, string port = RSF_PORT ); 
    int validateAddressFamily ( string ipaddr,  int *version );
    int validateHost ( string host, char **ip );

   /*
    * Cluster discovery
    * -----------------
    */ 
    int discoverClusters ( string address = "", int mode = rsf_discover_all, bool prettyPrint = false );

   /*
    * Data/request transmission functions
    * -----------------------------------
    */
    int readrsf (
      string ipaddr = DEFAULT_IP, string ipport = RSF_PORT, bool forceUpdate = false );

    int rsfcomms  (
      char *&buffer, string address = "", string port = "", string command = "",
      vector<string> authvec = vector<string>(), string data = "" );

    int serviceCommand (
      string address, string port,  string command, vector<string> authvec,
      unsigned serv_no, unsigned attemps = 1, string cmd_suffix = ""  );
  };

 /**
  * @brief      Class for the API internal information cache.
  */
  class RSFNetCache
  {
    RSFNetCache( const RSFNetCache& other );      // non construction-copyable
    RSFNetCache& operator=( const RSFNetCache& ); // non copyable

   /*
    * Networking variables
    * --------------------
    */ 
    struct    addrinfo hints;
    struct    addrinfo *res;
    char      ipaddr[INET6_ADDRSTRLEN];

   /*
    * Time variables
    * --------------
    */
    int       PollTime;         // Poll time between cache refreshes
    unsigned  LastUpdated;      // Time the cache was last updated
    unsigned  now;              // The current time

   /*
    * Local host variables
    * --------------------
    */ 
    string    LocalHostName;    // The hostname used for the localhost 
    string    localRequestNode; // Node to use for 'local' connections

   /*
    * Response packet handlers
    * ------------------------
    */ 
    char      getIDLetter ( rsf_info_letters id );
    unsigned  getLetterID ( char letter );

   /* 
    * Layout of rsfInfo object
    * ------------------------
    *
    * The rsfInfo object is where the cache info is stored. The layout of this
    * object is quite complex.
    * 
    * The position for the first layer selects which node info to use from the
    * cache.
    * 
    * The second postion selects which packet code (or letter) to obtain 
    * information for. This will usually be a value from the rsf_info_letters
    * enumerator.
    * 
    * The third postion selects the letter instance ( or line number ) to
    * obtain the information from. For example, if muliple lines may exist for
    * the same packet letter, this position selects which line to use.
    * 
    * The forth position selects the key to use from the letter line. This will
    * normally be a value from the letters corresponding rsfInfo letter values
    * enumerator. This will yield a string that will be the value assigned to
    * that letter key.
    * 
    * In summary:
    * 
    * Node selection   | vector<vector<vector<vector<string> > > >
    * Letter Selection | vector<vector<vector<string> > >
    * Letter ID Select | vector<vector<string> >
    * Letter Keys      | vector<string>
    * Value            | string
    *
    */
    vector<vector<vector<vector<string> > > > rsfInfo;

   /*
    * Cached nodes vector
    * -------------------
    */
    vector<string> rsfInfoNodes;

   /*
    * Local addresses list
    * --------------------
    */ 
    vector<string> getLocalAddrs();

   /*
    * Cache node info functions
    * -------------------------
    */  
    string _getInfoMachineName  ( vector<vector<vector<string> > > vvv  );
    string _getInfoHostName     ( vector<vector<vector<string> > > vvv  );

    bool _nodeInfoExists        ( string node );
    int _getNodePos             ( string node );

   /*
    * Cache-size functions
    * --------------------
    */
    size_t getCacheEndValue( unsigned id );

   /*
    * Info gathering function
    * -----------------------
    */ 
    int read_rsf(
      string ipaddr       = DEFAULT_IP,
      string ipport       = RSF_PORT,
      bool   force_update = false );

  public:
   /*
    * RSFNetCache Constructor
    */
    RSFNetCache();     
    
   /*
    * Localhost functions
    * -------------------
    */ 
    bool isLocalHost( string node );

   /*
    * Poll time functions
    * -------------------
    */ 
    int getPollTime( string ipaddr = DEFAULT_IP );

   /*
    * Cache update functions
    * ----------------------
    */
    int refreshCache (
      string ipaddr = DEFAULT_IP, string ipport = RSF_PORT, bool forceUpdate = false );
    int clearCache ();

   /*
    * Node enquiry functions
    * ----------------------
    */ 
    string addrHandler        ( string node );
    string getHostnameForNode ( string node );
    
   /*
    * Cache read functions
    * --------------------
    */ 
    vector<string> rsfInfoGetValueVector(
      string nde, rsf_info_letters flagType, unsigned lineNum );

    string rsfInfoGetValue(
      string nde, rsf_info_letters flagType, unsigned lineNum, unsigned value );
    
    unsigned getNumberOfLetterEntries(
      string nde, enum rsf_info_letters letter  );

   /*
    * Value type handlers
    * -------------------
    */
    string infoHandler (
      char letter, unsigned value, unsigned lineNum, string node, string _fn );

    char * infoCharHandler (
      char letter, unsigned value, unsigned lineNum, string node, string _fn );

    int infoIntHandler (
      char letter, unsigned value, unsigned lineNum, string node, string _fn );

    vector<string> infoVecHandler (
      char letter, unsigned value, unsigned lineNum, string node, string _fn );

    string getAllLetterInfoHandler (
      char letter, string node, bool nl, string _fn );

   /*
    * Make friends with the CLI class.
    */
    friend class CLI;
  };

 /**
  * @brief      Clas that provides functions to convert RSF response codes to
  *             user readable strings.
  */
  class RSFResponse
  {
  public:
    RSFResponse();     // RSFResponse Constructor
    ~RSFResponse();    // RSFResponse Destructor

    string getRSFResponseMsg   ( char str[1] );
    int    getRSFResponseStatus( char str[1] );
    string getRSFServiceState  ( int ssint  );
  };

 /** 
  * \brief Class for obtaining information for a cluster service.
  *
  *  Creates a service object from a service number, name, float name, description, and
  *  quad server vector.
  */
  class Services
  {
  private:

    vector<string> getQuadValues(
      enum rsfInfo_S_values quadValue,
      unsigned              sp,
      string                node,
      string                _fn );

    string getQuadElement(
      enum rsfInfo_S_values quadValue,
      unsigned              qn,
      unsigned              sp,
      string                node,
      string                _fn );

    int getQuadElementInt(
      enum rsfInfo_S_values quadValue,
      unsigned              qn,
      unsigned              sp,
      string                node,
      string                _fn );

    int isConfiguredForNode( int serviceNumber, string node );

  public:
    Services();     // Services Constructor
    ~Services();    // Services Destructor

   /*
    * Store counter function
    * ----------------------
    */ 
    unsigned getCount( string node = DEFAULT_IP );

   /*
    * Store access functions
    * ----------------------
    */ 
    int            getNumber     ( int sp = 0, string node = DEFAULT_IP );
    string         getName       ( int sp = 0, string node = DEFAULT_IP );
    string         getNetMask    ( int sp = 0, string node = DEFAULT_IP );
    string         getFloat      ( int sp = 0, string node = DEFAULT_IP );
    string         getDesc       ( int sp = 0, string node = DEFAULT_IP );
    vector<string> getQuads      ( int sp = 0, string node = DEFAULT_IP );
    int            getServerCount( int sp = 0, string node = DEFAULT_IP );

   /*
    * Repeated 'quad' access functions
    * --------------------------------
    */ 
    string getServer    ( int qn = 0, int sp = 0, string node = DEFAULT_IP );
    string getIPDevice  ( int qn = 0, int sp = 0, string node = DEFAULT_IP );
    int    getIniTimeout( int qn = 0, int sp = 0, string node = DEFAULT_IP );
    int    getRunTimeout( int qn = 0, int sp = 0, string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string getAllServInfo( string node = DEFAULT_IP, bool nl = true );

   /*
    * List functions
    * --------------
    */
    vector<string> getServicesList(
      string host = DEFAULT_IP, bool dynamicMode = false );

   /*
    * Enquiry functions
    * -----------------
    */   
    int getServiceNumber         ( string serviceName, string host = DEFAULT_IP );
    int isNodeConfiguredForName  ( string serviceName, string node );
    int isNodeConfiguredForNumber( int serviceNumber , string node );
  };


 /**
  * @brief      Class for obtaining the licensing status for RSF.
  */
  class VerLicInfo
  {
  public:
    VerLicInfo();             // VerLicInfo Constructor
    ~VerLicInfo();            // VerLicInfo Destructor

   /*
    * Read functions
    * --------------
    */ 
    string  getCliPro        ( string node = DEFAULT_IP );
    string  getRSFRel        ( string node = DEFAULT_IP );
    int     getResNode       ( string node = DEFAULT_IP );
    string  getLiExp         ( string node = DEFAULT_IP );
    string  getSysType       ( string node = DEFAULT_IP );
    string  getLicType       ( string node = DEFAULT_IP );
    string  getMacID         ( string node = DEFAULT_IP );
    int     getLicCheckResult( string node = DEFAULT_IP );
    int     getDemoStat      ( string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string  getAllVerLicInfo( string node = DEFAULT_IP, bool nl = true );
  };

 /**
  * @brief      Class for obtaining information for the currently connected RSF
  *             node.
  * 
  * Returns the machine name, hostname and IP address of the current connected
  * node.
  */
  class WhichNode
  {
  public:
    WhichNode();              // WhichNode Constructor
    ~WhichNode();             // WhichNode Destructor

   /*
    * Read functions
    * --------------
    */ 
    string getMachineName  ( string node = DEFAULT_IP );
    string getHostName     ( string node = DEFAULT_IP );
    string getHostID       ( string node = DEFAULT_IP );
    string getHostIP       ( string node = DEFAULT_IP );
    string getAddressFamily( string node = DEFAULT_IP );

   /*
    * All info function
    * -----------------
    */ 
    string getAllReleaseInfo( string node = DEFAULT_IP, bool nl = true );

   /*
    * Enquiry functions
    * -----------------
    */
    int isLocalNode( string node );
  };

 /**
  * @brief      Class for providing the Command Line Interface.
  */
  class CLI
  {
  private:  
    CLI( const CLI& other );      // non construction-copyable
    CLI& operator=( const CLI& ); // non copyable

    /* Internal CLI variables */

    /* Booleans */
    bool verboseFlag;     // Set's verbose mode for the CLI
    bool silentFlag;      // Set's silent mode for the CLI
    bool localFlag;       // Restricts connections to the local node only
    bool ipFlag;          // Displays's hostname as an IP address if true
    bool interactiveFlag; // Allows interactive input if true
    bool stickyFlag;      // Enables sticky move for the move command
    bool permFlag;        // Sets licence type for licmsg to permanent

    /* Integers */
    int attemptCount;     // The target number of attempts for the service move

    /* Strings */
    string _hostIP;       // The target node for RSF connections 
    string _initIP;       // The node that should be contacted first
    string _targetSvc;    // The target service for the command passed to the CLI 
    string _floatName;    // The floating name to check when using vname command
    string _logMsg;       // The custom debug message to be written to the log file

    string _authUser;     // The username passed to the CLI for authentication
    string _authPass;     // The password passed to the CLI for authentication

    vector<string> _authObject;


    string getPriorityString( int priority );
    string getUnblockedString( int status );

   /*
    * List functions
    * --------------
    */
    void printInterfacesList(
      bool   verbose   = false,
      string host      = "",
      string service   = "",
      string commsHost = DEFAULT_IP );

    void printNodesList(
      bool   verbose  = false,
      string host     = DEFAULT_IP,
      bool   tagLocal = false );

    void printServicesList(
      bool   verbose   = false,
      string host      = "",
      string service   = "",
      string commsHost = DEFAULT_IP );

    void printHeartbeatsList(
      bool   verbose   = false,
      string host      = "",
      string commsHost = DEFAULT_IP );



    void printServiceModes ( string host = DEFAULT_IP );
    int  printServiceStatus( string host = DEFAULT_IP );


    /* Functions */
    int  validateClusterCRC  ( string host );
    void printHeartbeatState ( string host, bool listMode = false );
    void printNodeState      ( string host );
    void printServiceState   ( string host );
    void printServiceConfig  ( string host );


    int printLicenceMsg(
      bool           reqPerm    = true,
      string         srcNode    = DEFAULT_IP,
      vector<string> hostIDList = vector<string>(),
      string         custref    = "rsfapi",
      string         ponum      = "N/A" );
    
    int _initialIPChecks       ( string& host );
    int getPID                  ();
    int getAuthCredentials     ( string host );

    int testConnection(
      char *         &rsfbuf,
      string         address,
      string         port,
      vector<string> authObj    = vector<string>(),
      bool           showErrors = true );

   /*
    * CLI action command wrapper
    * --------------------------
    */ 
    int serviceAction(
      int (RSFActions::*fptr)(
        string, string, unsigned, vector<string>, string, string ),
      RSFActions& c,
      string host,
      unsigned argc,
      char *argv[] );

    int serviceMove( string commsHost, int quickmove, int argCount, char *argv[] );

   /*
    * Help functions
    * --------------
    */ 
    void printCmdHelp       ( int i );
    void printCmdSyntax    ( int s );
    void printSwitchesHelp  ( int i );
    int  printFullCmdHelp         ( string str );
    void printHelpList        ( );

   /*
    * Input validation functions
    * --------------------------
    */
    char * processFlag ( char *optarg );
    int    handleFlags ( int argc, char* argv[] );

  public:
    CLI();         // CLI Constructor

   /*
    * CLI main function
    * -----------------
    */ 
    int launch( int argc, char* argv[] );
  };

  extern Auth         rsfAuth;
  extern Backup       rsfBackup;
  extern CLI          rsfCLI;
  extern Cluster      rsfCluster;
  extern Debug        rsfDebug;
  extern ExtraNetInfo rsfExtraNet;
  extern Flags        rsfFlags;
  extern HBConfig     rsfHBConfig;
  extern Instances    rsfInstances;
  extern Node         rsfNode;
  extern Release      rsfRelease;
  extern RSFActions   rsfActions;
  extern RSFNet       rsfNet;
  extern RSFNetCache  rsfCache;
  extern RSFResponse  rsfResponse;
  extern Services     rsfServices;
  extern VerLicInfo   rsfLicence;
  extern WhichNode    rsfLocalNode;
}


/* =============================================================================
** API Utility Functions
** ---------------------
** 
** Function declarations for the global API functions. These will be exposed
** to the rsfUtils namespace to avoid 'polluting' the std namespace and to
** reduce the risk of a function name conflict with existing project names.
**
** =============================================================================
*/
namespace rsfUtils
{
  using namespace std;

  /* =============================================================================
  ** Function Templates
  ** ------------------
  **
  **  Template functions are defined here. Since templates require the 'pattern'
  **  and the 'definition' to be declared together AND as this function is
  **  intended to be used by functions contained in other source files, the
  **  whole template function is contained here.
  **
  ** =============================================================================
  */
  template<typename T, typename A>
  string vectostr( std::vector<T,A> const& v )
  {
    std::stringstream ss;

    for ( size_t i = 0; i < v.size(); i++ )
    {
      if ( i != 0 ) { ss << ", " ; }
      ss << v[i];
    }
    return ss.str();
  }

 /*
  * File Functions
  * --------------
  */
  bool fileExists( const std::string& name );

 /*
  * Type Conversion Functions
  * -------------------------
  */
  char * intToChar  ( int num );
  string intToString( int num, unsigned padding = 0 );

  char * stringToChar( string str );
  int    stringToInt ( string str );

 /*
  * Formatting Functions
  * --------------------
  */ 
  string formatStr( string  strin, int minSize = 10, bool alignLeft = true );
  string formatInt( int intin, int minSize = 10, bool alignLeft = true );

 /*
  * Time & Date Functions
  * ---------------------
  */
  string getTimeStr( time_t tin = 0 );

 /*
  * String splitters
  * ----------------
  */
  void tokenizer(
    const string& str, vector<string>& tok, const string& del = " " );

  vector<string> newlineDel( const string& str );

 /*
  * Info printer function
  * ---------------------
  */
  void printClassMsg( string name, bool destruct );

 /*
  * Type check functions
  * --------------------
  */ 
  bool isNotDigit( char c );

 /*
  * Transformation Functions
  * ------------------------
  */
  string lowerStr( string s );

 /*
  * Vector functions
  * ----------------
  */
  bool vectorHasValue( string value, vector<string> v );

} // namespace rsfUtils

/* =============================================================================
** SHA1 Library
** ------------
** 
** SHA1 library used in this API. The copyright notice for the SHA1 library
** follows:
**  --------------------------------------------------------------------------
**
**  Copyright (C) 1998, 2009
**  Paul E. Jones <paulej@packetizer.com>
**  All Rights Reserved.
**
******************************************************************************
**  $Id$
******************************************************************************
**
**  Description:
**      This class implements the Secure Hashing Standard as defined
**      in FIPS PUB 180-1 published April 17, 1995.
**
**      The Secure Hashing Standard, which uses the Secure Hashing
**      Algorithm (SHA), produces a 160-bit message digest for a
**      given data stream.  In theory, it is highly improbable that
**      two messages will produce the same message digest.  Therefore,
**      this algorithm can serve as a means of providing a "fingerprint"
**      for a message.
**
**  Portability Issues:
**      SHA-1 is defined in terms of 32-bit "words".  This code was
**      written with the expectation that the processor has at least
**      a 32-bit machine word size.  If the machine word size is larger,
**      the code should still function properly.  One caveat to that
**      is that the input functions taking characters and character arrays
**      assume that only 8 bits of information are stored in each character.
**
**  Caveats:
**      SHA-1 is designed to work with messages less than 2^64 bits long.
**      Although SHA-1 allows a message digest to be generated for
**      messages of any number of bits less than 2^64, this implementation
**      only works with messages with a length that is a multiple of 8
**      bits.
**
** =============================================================================
*/
class SHA1
{
public:
  SHA1();
  virtual ~SHA1();

  // Re-initialize the class
  void Reset();

  // Returns the message digest
  bool Result(unsigned *message_digest_array);

  void Input( const unsigned char *message_array, unsigned length);
  void Input( const char  *message_array, unsigned length);
  void Input(unsigned char message_element);
  void Input(char message_element);

  SHA1& operator<<(const char *message_array);
  SHA1& operator<<(const unsigned char *message_array);
  SHA1& operator<<(const char message_element);
  SHA1& operator<<(const unsigned char message_element);

private:
  // Process the next 512 bits of the message
  void ProcessMessageBlock();

  // Pads the current message block to 512 bits
  void PadMessage();

  // Performs a circular left shift operation
  inline unsigned CircularShift(int bits, unsigned word);

  unsigned H[5];                      // Message digest buffers
  unsigned Length_Low;                // Message length in bits
  unsigned Length_High;               // Message length in bits
  unsigned char Message_Block[64];    // 512-bit message blocks
  int Message_Block_Index;            // Index into message block array
  bool Computed;                      // Is the digest computed?
  bool Corrupted;                     // Is the message digest corruped?
};

// Need to undefine C (which is defined in the rsflib header files) for the
// SHA-1 library to compile correctly
#undef C


#endif /* LIBRSFAPI_H_ */
