/* =============================================================================
**
**  File        : librsfcdb.h
**  Copyright   : High-Availability.com Ltd (c) 1995-2020
**
** =============================================================================
*/
/**
 * @file  librsfcdb.h
 * @brief Header file for RSFCDB
 */
#ifndef LIBRSFCDB_H_
#define LIBRSFCDB_H_

#include <algorithm>
#include <cstring>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <map>
#include <sstream>
#include <string>
#include <vector>
#include <cstdlib>
#include "errno.h"
#include "getopt.h"
#include "unistd.h"

#include <sys/stat.h>
#include <stdio.h>
#include <time.h>
#include <utime.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <fcntl.h>
#include <limits.h>

/*
** Enable 'traceback' printing for a lock if using Solaris 10+
*/
#if defined(_DEBUG_SOLARIS)
#include <procfs.h>
#endif

#if defined(__FreeBSD__) || defined(linux) || defined(__linux__)
#include <sys/wait.h>
#endif

#include "sqlite3.h"

#include "rsfcon.h"


/*
** RSFCDB Version Number
*/
#define CDB_VERSION_NUMBER       "3.1.6"

/*
** Default save path for generated bootstrap configuration files
*/
#define CDB_BOOTSTRAP_PATH       "config"

/*
** Default path for the settings file
*/
#define CDB_SETTINGS_FILE        "/opt/HAC/RSF-1/etc/.cluster_db.cfg"

/*
** Default path for the database file
*/
#define CDB_DEFAULT_DB_PATH      "rsfcdb.db"

/*
** Default path for the generated config file.
*/
#define CDB_DEFAULT_CFG_PATH     "config"

/*
** Default logical expression for IP and FC devices
*/
#define CDB_DEFAULT_IPD_LOGIC     "AND"
#define CDB_DEFAULT_FCD_LOGIC     "AND"

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)

#ifndef CDB_DEL_STR
#define CDB_DEL_STR "#"
#endif

#ifndef CDB_DEL_CHAR
#define CDB_DEL_CHAR '#'
#endif

#ifndef CDB_MON_DEL_STR
#define CDB_MON_DEL_STR ","
#endif

#ifndef CDB_MON_DEL_CHAR
#define CDB_MON_DEL_CHAR ','
#endif

/*
** Prefixes to use for heartbeat and server node entries in the config file
*/
#define CDB_HBT_NODE_PREFIX   "%"
#define CDB_SVR_NODE_PREFIX   "%"

/*
** Default value to use for a service VIP if none is provided
*/
#define CDB_DEFAULT_VIP     "NONE"

/*
** Reserved charater to specify a heartbeat with no defined via path
*/
#define CDB_NO_VIA_PATH     "-"

/*
** Default value to use for the service storage type if none is provided
*/
#define CDB_DEFAULT_STOR_TYPE "disk"

/*
** Database property keys
*/
#define CDB_REOPEN_SOCK_PROP "prop_nethb_reopensocket"
#define CDB_IP_MON_PROP      "prop_enable_ip_monitor"
#define CDB_FC_MON_PROP      "prop_enable_fc_monitor"

/*
** SQLite specific values
*/
#define CDB_SQLITE_NULL     "[NULL]"

/* =============================================================================
** Debug File Preprocessing
** ------------------------
*/
#define CDB_ENABLE_DEBUG  "/tmp/rsfcdb-debug-on"
#define CDB_DEBUG_FILE    "/opt/HAC/RSF-1/log/rsfcdb.log"

/* =============================================================================
** Log Printing
** ------------
*/
#ifndef CDBLOG
#define CDBLOG(X) rsfcdb::cdblog.write ( std::ostringstream().flush() << X, false )
#endif

#ifndef CDBLOGHIGH
#define CDBLOGHIGH(X) rsfcdb::cdblog.write ( std::ostringstream().flush() << X, true )
#endif

#ifndef CDBERR
#define CDBERR(X,Y) rsfcdb::cdblog.error ( X, std::ostringstream().flush() << Y )
#endif

#ifndef CDBVERBOSE
#define CDBVERBOSE(X,Y) rsfcdb::cdblog.verbose ( X, std::ostringstream().flush() << Y )
#endif


/* =============================================================================
** Debug/Logging Triggers
** ----------------------
*/
#ifndef CDB_TRIGGER_ERRPRINT
#define CDB_TRIGGER_ERRPRINT(X) rsfcdb::cdblog.trigger_error_print(X);
#endif

#ifndef CDB_TRIGGER_ERRLOG
#define CDB_TRIGGER_ERRLOG(X)   rsfcdb::cdblog.trigger_error_logging(X);
#endif

/* =============================================================================
** Debug/Logging Status
** --------------------
*/
#define CDB_STAT_VERBOSE      rsfcdb::cdblog.get_verbose_status()
#define CDB_STAT_ERRPRINT     rsfcdb::cdblog.get_error_print_status()
#define CDB_STAT_ERRLOG       rsfcdb::cdblog.get_error_logging_status()
#define CDB_STAT_CONFIG_DEBUG rsfcdb::cdblog.get_config_debug_status()
#define CDB_LOG_BREAK         rsfcdb::cdblog.get_log_break()
#define CDB_LOG_SUBBREAK      rsfcdb::cdblog.get_log_subbreak()
#define CDB_ERROR             rsfcdb::cdblog.get_err()

/* =============================================================================
** Console Formatting
** ------------------
*/
#ifndef SETCOL
#define SETCOL(X) "\033[" << X << "m"
#endif

/* =============================================================================
** Reservered API Keys
** ------------------
**  Define some reserverd keys in the pre-processor. This makes it easier to
**  manage should any of the keys need to be changed in future.
*/
#define CDB_GUID_KEY        "ZPOOL_GUID"
#define CDB_COMMENT_KEY     "HEADER_COMMENT"
#define CDB_SINGLE_NODE_KEY "SINGLE_NODE"
#define CDB_FC_LOGIC_KEY    "FC_LOGIC"
#define CDB_FC_MONITOR_TAG  "%_FCDEV_MONITOR"

/**
 *  @enum cdb_verbose_levels
 * 
 *  Enumerator for the valid levels of verbosity supported by rsfcdb.
 *
 *  CDB_VERBOSE_OFF:  No verbose output is printed.
 *
 *  CDB_VERBOSE_LOW:  Verbose info is printed for results of significant
 *                    internal checks/comparisions performed by the function
 *                    called.
 *
 *  CDB_VERBOSE_MED:  As CDB_VERBOSE_LOW but output is also printed to inform
 *                    about check's that a are about to be called.
 *
 *  CDB_VERBOSE_HIGH: All verbose output is printed.
 */
enum cdb_verbose_levels
{
  CDB_VERBOSE_OFF   = 0,
  CDB_VERBOSE_LOW   = 1,
  CDB_VERBOSE_MED   = 2,
  CDB_VERBOSE_HIGH  = 3
};

/**
 * @enum        cdb_path_type
 * @brief       Enumerator for the supported file types.
 *
 *              FILETYPE     : A file
 *              DIRTYPE      : A directory/folder.
 *              UNKNOWNTYPE  : Other unknown file type.
 */
enum cdb_path_type
{
  FILETYPE    =  0,
  DIRTYPE     =  1,
  UNKNOWNTYPE = -1
};

/**
 * @enum        cdb_error_enum
 * 
 * @brief       The error code enumerated list
 */
enum cdb_error_enum
{
    CDB_OK = 0,         /*!< OK return code */

    // CLI specific
    //
    CDB_CLI_BATCH,      /**< Error(s) occurred during batch command */
    CDB_CLI_CMDUNKNOWN, /**< Cannot find the requested command */
    CDB_CLI_EMPTY,      /**< Empty vector returned as value */
    CDB_CLI_NOARG,      /**< No argument(s) provided to command */
    CDB_CLI_NOCMD,      /**< No command provided */
    CDB_CLI_NODBFLAGS,  /**< No database path and/or type provided */
    CDB_CLI_NODBTYPE,   /**< No database type provided */
    CDB_CLI_NODBPATH,   /**< No database path provided */
    CDB_CLI_NOFLAGVAL,  /**< No value provided to CLI flag */

    // Database check
    //
    CDB_UNKNOWN_DB,     /**< Unknown database type selected */
    CDB_UNSUPPORTEDDB,  /**< Command unsupported for current DB type */

    // I/O errors
    //
    CDB_FILE_CLOSE,     /**< Failed to close file */
    CDB_FILE_NOTFOUND,  /**< Failed to locate the file */
    CDB_FILE_OPEN,      /**< Failed to open file */
    CDB_FILE_REMOVE,    /**< Failed to remove file */
    CDB_FILE_RENAME,    /**< Failed to rename file */
    CDB_FILE_TOUCH,     /**< Failed to touch file */
    CDB_FILE_WRITE,     /**< Failed to write to a file */

    // SQLite specific
    //
    CDB_SQL_CLOSE,      /**< Failed to close SQLite database. */
    CDB_SQL_CLOSEDDB,   /**< Attempted operation on closed SQLite database */
    CDB_SQL_CREATEDB,   /**< Cannot create new SQLite database */
    CDB_SQL_DELETE,     /**< Failed to remove entry from SQLite database */
    CDB_SQL_INVALIDDB,  /**< File is not a valid SQLite database */
    CDB_SQL_LOCKED,     /**< Failed to obtain lock for SQLite database */
    CDB_SQL_NOCHANGE,   /**< SQLite command made no changes to database */
    CDB_SQL_OPEN,       /**< Failed to open SQLite database */
    CDB_SQL_UPDATE,     /**< Failed to update entry in SQLite database */
    CDB_SQL_WRITE,      /**< Failed to write entry to SQLite database */
    CDB_SQL_EXEC,       /**< SQLite command failed */
    CDB_SQL_COLUMN,     /**< SQLite table column not found */
    CDB_SQL_BACKUP,     /**< Failed to initialise/restore snapshot */
  
    // Key/Value Error
    //
    CDB_DESCPROPSONLY,  /**< Descriptions reserved for 'prop' values only */
    CDB_EMPTY_STRING,   /**< Empty string detected/provided to value */
    CDB_HBT_TYPE,       /**< Unknown heartbeat type */
    CDB_INVALIDCHAR,    /**< (Too many) invalid characters in string */
    CDB_RESERVEDKEY,    /**< Key name is reserved for an internal value */
    CDB_STRINGNOTINT,   /**< String contains non-numerical characters */
    CDB_VALCONFLICT,    /**< Value conflicts with other value provided */
    CDB_VALEXISTS,      /**< Value already exists in database */
    CDB_VALNOTFOUND,    /**< Requested value not found in database */
    CDB_VALOUTOFRANGE,  /**< Value is not within the permitted range */
    CDB_VALTOOHIGH,     /**< Value is too high */
    CDB_VALTOOLOW,      /**< Value is too low */

    // Networking
    //
    CDB_GETADDRINFO,    /**< Error reported by getaddrinfo() */
    CDB_ADDRFAMILY,     /**< IP is of an unknown address family */

    // Not Available
    //
    CDB_NOSERVER,       /**< Server/node not found in database */
    CDB_NOSERVICE,      /**< Service not found in database */
    CDB_NOVIP,          /**< Vip not found for service */
    CDB_PROPNOTFOUND,   /**< Property not found in database */
    CDB_DEPRECATED,     /**< Function has depecated */

    // Service Lines & internal errors
    //
    CDB_VALIDFAIL,      /**< Internal validation check failed */
    CDB_STRTOINTFAIL,   /**< Failed to convert string to integer */

    // Vanilla Status
    //
    CDB_NOTVAN,         /**< Database is not 'vanilla' */
    CDB_UNSETVAN,       /**< Failed to unset vanilla flag from database */

    // Vector related
    //
    CDB_INVALIDTOKENS,  /**< An invalid amount of tokens were provided */
    CDB_TOOFEWTOKENS,   /**< Not enough 'tokens' found */
    CDB_TOOMANYTOKENS,  /**< Too many 'tokens' found */

    // IP Address Family
    //
    CDB_INET_IPV4,      /**< Address family is IPv4 */
    CDB_INET_IPV6,      /**< Address family is IPv6 */

    CDB_ERR_COUNT       /**/
};

/**
 * @enum          db_selection
 *
 * @brief         Enumerator of supported databases. Although only SQLite3 is
 *                supported at this time, this may change in the future and is
 *                provided to ease the migration to a different database type.
 */
enum db_selection
{
  SQLITE3 = 0
};

/**
 * @enum          heartbeat_type
 * 
 * @brief         Enumerator of valid heartbeat types.
 */
enum heartbeat_type
{
  HBT_NET = 0,  /*  Legacy network heartbeat      */
  HBT_NET4,     /*  IPv4 network heartbeat        */
  HBT_NET6,     /*  IPv6 network heartbeat        */
  HBT_SERIAL,   /*  Serial heartbeat              */
  HBT_DISC,     /*  Disc heartbeat                */
  HBT_ANY       /*  Any heartbeat type            */
};

/**
 * @enum          primary_vip_filter
 * 
 * @brief         Enumerator of valid primary vip filter levels.
 */
enum primary_vip_filter
{
  VIP_ALL = 0,    /* Obtain all vips */
  VIP_PRIMARY,    /* Obtain only primary vips */
  VIP_ADDITIONAL  /* Obtain only additional vips */
};

/**
 * @enum        cdb_ansi_colour_codes
 *
 * @brief       Colour codes for formatted ANSI console output.
 */
enum cdb_ansi_colour_codes
{
  CDB_FG_BLACK    = 30,
  CDB_FG_RED      = 31,
  CDB_FG_GREEN    = 32,
  CDB_FG_YELLOW   = 33,
  CDB_FG_BLUE     = 34,
  CDB_FG_MAGENTA  = 35,
  CDB_FG_CYAN     = 36,
  CDB_FG_WHITE    = 37,
  CDB_FG_DEFAULT  = 39,
  CDB_BG_BLACK    = 40,
  CDB_BG_RED      = 41,
  CDB_BG_GREEN    = 42,
  CDB_BG_YELLOW   = 43,
  CDB_BG_BLUE     = 44,
  CDB_BG_MAGENTA  = 45,
  CDB_BG_CYAN     = 46,
  CDB_BG_WHITE    = 47,
  CDB_BG_DEFAULT  = 49
};

/* =============================================================================
** RSFCDB Namespace
** ----------------
** =============================================================================
*/
namespace rsfcdb
{
  using std::string;
  using std::cout;
  using std::endl;
  using std::vector;

 /**
  * @fn         cdb_error_string(int err)
  *
  * @brief      Returns a human readable string explaining the meaning of an
  *             rsfcdb return code.
  *
  * @param      err   The return code to obtain a string for.
  *
  * @return     Non-empty string containing the error code meaning.
  *             String contained in square brackets on error.
  */
  const char * cdb_error_string( int err );


 /* ==========================================================================
  *
  *   Class     : RSFDB_Prop
  *   Instance  : rsfdb_prop
  *
  * ==========================================================================
  */
#define PROP_TIME_UNKNOWN   "<unknown>"

#define PROP_KEY_COL        "key"
#define PROP_VAL_COL        "value"
#define PROP_DESC_COL       "description"
#define PROP_CTIME_COL      "creation_time"
#define PROP_UTIME_COL      "update_time"
#define PROP_REF_COL        "update_reference"
#define PROP_PVT_COL        "is_private"

 /**
  * @brief Provides CRUD (Create, Remove, Update & Delete) functionality for custom
  * key/values in the database.
  */
  class RSFDB_Prop
  {
    // Friend declaration to allow prop line generation in private CLI class
    // functions and for control of the vanilla mode status.
    //
    friend class CDB_CLI;

    // Friend declaration to allow the SQLite Heartbeat class to control the
    // single node mode status.
    //
    friend class SQLite_Heartbeat;

    // Friend declaration to allow the SQLite Services class to control the
    // logic property for the FC Device monitor.
    //
    friend class SQLite_Services;

    // Friend declaration to allow the Config class to control the comment header.
    //
    friend class RSFDB_Config;

    // Allow GUI Users access to the private timestamp functions.
    //
    friend class GUI_Users;

  private:
    bool key_is_reserved( string key );
    
    int perform_key_actions( string  db, string  key, string  value );

    int perform_sql_cmd_prop(
      string  dbName,
      string  sqlCmd,
      int     * changeCount = NULL,
      int     vanilla       = 0 );

    string sqlite_read_prop_column( string dbName, string prop, string column );

   /*
    * List of reserved internal keys
    * ------------------------------
    */
    static const char* config_key_list[];

   /*
    *  Property timestamp_description commands
    *  ---------------------------------------
    */
    string get_timestamp();

    string format_description_string( string str );

   /*
    *  Root property CRUD commands
    *  ---------------------------
    */
    int root_create(
      string  dbName,
      string  key,
      string  value,
      string  ref       = "",
      bool    isPrivate = false );
    
    string root_read(
      string  dbName,
      string  key,
      bool    isPrivate = false );
    
    int root_update(
      string  dbName,
      string  key,
      string  newval,
      string  ref       = "",
      bool    isPrivate = false );
    
    int root_delete(
      string dbName,
      string key,
      bool   isPrivate = false );

   /*
    *  Prop line functions
    *  -------------------
    */
    string generate_prop_line(
      string  dbName,
      string  key,
      int     valFillLen  = 10,
      bool    listMode    = false,
      bool    showDesc    = false,
      bool    showTS      = false );  

   /*
    *  Vanilla status commands commands
    *  --------------------------------
    */
    bool vanilla_status( string dbName );
    int  make_vanilla  ( string dbName );

  public:
   /*
    *  Class (de-)constructors
    *  -----------------------
    */
    RSFDB_Prop();
    ~RSFDB_Prop();

   /*
    * Property key list
    * -----------------
    */ 
    vector<string> list_custom_prop_keys( string dbName );

   /*
    *  Property C.R.U.D commands
    *  -------------------------  
    */
    int    create_prop( string dbName, string key, string value, string ref = "" );
    string read_prop  ( string dbName, string key );
    int    update_prop( string dbName, string key, string newval, string ref = "" );
    int    delete_prop( string dbName, string key );
    
    int check_key_exists( string dbName, string key );

   /*
    *  Read property reference
    *  -----------------------
    */
    string get_prop_reference( string dbName, string prop );

   /*
    *  Timestamp commands
    *  ------------------
    */
    int get_prop_creation_time( string dbName, string key );
    int get_prop_update_time  ( string dbName, string key );

   /*
    *  Timestamp description commands
    *  ------------------------------
    */
    int    set_prop_description( string dbName, string key, string desc );
    string get_prop_description( string dbName, string key );
  };

  extern RSFDB_Prop rsfdb_prop;


 /* ==========================================================================
  *
  *   Class     : RSFDB_Config
  *   Instance  : config
  *
  * ==========================================================================
  */

 /**
  * @brief Contains functions that generate the config file and allow a current
  *        config file to be imported to a database.
  */
  class RSFDB_Config
  {
  private:

   /*
    * Snapshot handler
    * ----------------
    */
    int _handle_snapshot_response( int res );

   /*
    * Config file printing functions
    * ------------------------------
    */
    string generate_monitor_tag ( string service, string server, string interface );

    void print_config_line(
      string  key,
      string  value,
      bool    addPrefix,
      bool    quoted,
      bool    debug );

    int print_global_section( string db );
    
    void print_machine_heartbeats(
      string  dbName,
      string  srcHost,
      string  srcIP = "" );

    void print_sorted_services_section(
      string         dbName,
      vector<string> svcVec,
      bool           genMonTags,
      bool           usePrimary );
    
    void print_service_config_debug(
      string  key,
      string  value,
      string  service,
      bool    envPrefix,
      bool    quoteValue,
      bool    debug );
    
    int print_services_section(
      string dbName, string serivce, bool genMonTags, bool usePrimary );

   /*
    * Config file to database functions
    * --------------------------------- 
    */
    string get_config_key_alias( string key );

    void handle_config_to_db_verbose_found(
      string line,
      string key   = "" ,
      string value = "" );

    void handle_config_to_db_verbose_add(
      string key,
      string value );

    void handle_config_to_db_verbose_result(
      int       ret,
      string    key,
      unsigned *errCnt );

    void handle_config_to_db_heartbeat(
      string          dbName,
      vector<string>  line_vec,
      string          srcHbtStr,
      heartbeat_type  hbtType,
      unsigned        *errCnt );

    void handle_config_to_db_service_msg(
      string svcStr,
      string keyStr,
      string valStr,
      string svrStr = "");

    void handle_config_to_db_global_msg( string keyStr, string valStr );

    int config_to_db_global(
      string    dbName,
      string    keyStr,
      string    valStr,
      unsigned  *errCnt );

    int config_to_db_heartbeat(
      string          dbName,
      heartbeat_type  hbtType,
      string          srcHbtStr,
      string          dstHbtStr,
      string          pathStr     = "",
      int             readBlock   = 0,
      int             writeBlock  = 0,
      string          tagStr      = "",
      unsigned        *errCnt     = NULL );

   /*
    * Import database functions
    * -------------------------
    */
    vector<string> make_pool_cache_object( string svcStr, string poolStr, string guidStr );

    int import_sqlite_database_global    ( string outputDB, string inputDB );
    int import_sqlite_database_machine   ( string outputDB, string inputDB );
    int import_sqlite_database_service   ( string outputDB, string inputDB );
    int import_sqlite_database_server    ( string outputDB, string inputDB );
    int import_sqlite_database_properties( string outputDB, string inputDB );

    int import_sqlite_database           ( string outputDB, string inputDB );
    int apply_sqlite_database            ( string outputDB, string tmpDB   );

    int make_hash_file                   ( string dbName, string outputFile );

  public:
   /*
    *  Class (de-)constructors
    *  -----------------------
    */
    RSFDB_Config();
    ~RSFDB_Config();

   /*
    * Snapshot commands
    * -----------------
    */
    int create_sql_snapshot ( string dbName, string snapshotName );
    int restore_sql_snapshot( string dbName, string snapshotName, bool cleanup = false );
    int delete_sql_snapshot ( string snapshotName );
  
   /*
    * Config file commands
    * --------------------
    */ 
    int config_preview( string dbName, bool genMonTags = false, bool usePrimary = false );
    
    int make_config(
      string  dbName,
      string  cfgPath     = CDB_DEFAULT_CFG_PATH,
      bool    bootstrap   = false,
      bool    monitorTags = false,
      bool    usePrimary  = false );

    int config_to_db( string cfgPath, string dbName );

   /*
    * Database reset actions
    * ----------------------
    */
    int reset_config_db  ( string dbName );
    int reset_rsf_cluster( string dbName );

    int rename_cluster_node (
      string dbName, string oldNode, string newNode, vector< vector< string> > privHbts );
    int delete_cluster_node (string dbName, string node);

   /*
    *  Header comment commands
    *  -----------------------
    */
    string  read_header_comment  ( string dbName );
    int     create_header_comment( string dbName, string comment );
    int     update_header_comment( string dbName, string comment );
    int     delete_header_comment( string dbName );
    
    friend class CDB_CLI;
  };

  extern RSFDB_Config rsfdb_config;

 /* ==========================================================================
  *
  *   Class     : RSFDB_Global
  *   Instance  : rsfdb_global
  *
  * ==========================================================================
  */

 /**
  * @brief Contains functions to configure global values in the selected
  *        database.
  */
  class RSFDB_Global
  {
  private:

  public:
   /*
    *  Class (de-)constructors
    *  -----------------------
    */
    RSFDB_Global();
    ~RSFDB_Global();

   /*
    *  Cluster Name
    *  ------------
    */
    string  read_cluster_name  ( string db );
    int     create_cluster_name( string db, string name );
    int     update_cluster_name( string db, string name );
    int     delete_cluster_name( string db );

   /*
    *  Cluster Description
    *  -------------------
    */
    string  read_cluster_description  ( string db );
    int     create_cluster_description( string db, string desc );
    int     update_cluster_description( string db, string desc );
    int     delete_cluster_description( string db );

   /*
    *  Cluster Identifier
    *  ------------------
    */
    string  read_cluster_identifier  ( string db );
    int     create_cluster_identifier( string db, string ident );
    int     update_cluster_identifier( string db, string ident );
    int     delete_cluster_identifier( string db );

   /*
    *  Disc Heartbeat Backoff
    *  ----------------------
    */
    string  read_disc_hb_backoff  ( string db );
    int     create_disc_hb_backoff( string db, int start, int multiplier, int max );
    int     update_disc_hb_backoff( string db, int start, int multiplier, int max );
    int     delete_disc_hb_backoff( string db );

   /*
    *  Poll Time
    *  ---------
    */
    string  read_poll_time  ( string db );
    int     create_poll_time( string db, int pt );
    int     update_poll_time( string db, int pt );
    int     delete_poll_time( string db );

   /*
    *  Realtime
    *  --------
    */
    string  read_realtime  ( string db );
    int     create_realtime( string db, int rt );
    int     update_realtime( string db, int rt );
    int     delete_realtime( string db );

   /*
    *  Event Notifier
    *  --------------
    */
    string  read_event_notify  ( string db );
    int     create_event_notify( string db, string en  );
    int     update_event_notify( string db, string en  );
    int     delete_event_notify( string db );

   /*
    *  IP Device Monitor
    *  -----------------
    */
    string  read_ipdevice_monitor  ( string db );
    int     create_ipdevice_monitor( string db, int pu, int pd );
    int     update_ipdevice_monitor( string db, int pu, int pd );
    int     delete_ipdevice_monitor( string db );

   /*
    *  FC Device Monitor
    *  -----------------
    */
    string  read_fcdevice_monitor  ( string db );
    int     create_fcdevice_monitor( string db, int pu, int pd );
    int     update_fcdevice_monitor( string db, int pu, int pd );
    int     delete_fcdevice_monitor( string db );

   /*
    *  Heartbeat Timeout
    *  -----------------
    */
    string  read_heartbeat_timeout  ( string db );
    int     create_heartbeat_timeout( string db, int ht );
    int     update_heartbeat_timeout( string db, int ht );
    int     delete_heartbeat_timeout( string db );
  };

  extern RSFDB_Global rsfdb_global;

 /* ==========================================================================
  *
  *   Class     : RSFDB_Heartbeat
  *   Instance  : rsfdb_heartbeat
  *
  * ==========================================================================
  */

 /**
  * @brief Contains functions to configure heartbeat values in the
  *        selected database.
  */
  class RSFDB_Heartbeats
  {
  private:
   /*
    * Private Heartbeat 'batch' Commands
    * ----------------------------------
    */
    vector<vector<string> > _list_heartbeats(
      string         db,
      heartbeat_type type,
      string         srcNode = "" );
    
    int _delete_all_heartbeats(
      string          db,
      heartbeat_type  type,
      string          srcNode = "" );

  public:
   /*
    * Class (de-)constructors
    * -----------------------
    */ 
    RSFDB_Heartbeats();
    ~RSFDB_Heartbeats();

    vector<string> get_nodes_list( string dbName, bool format = false );

   /*
    * Read Heartbeats
    * ---------------
    */
    vector<string> get_net_config_lines   ( string dbName, string srcHost );
    vector<string> get_serial_config_lines( string dbName, string srcHost );
    vector<string> get_disc_config_lines  ( string dbName, string srcHost );

   /*
    * Add Heartbeats
    * --------------
    */
    int create_net_heartbeat(
      string dbName,
      string srcNode,
      string dstNode,
      string viaPath = "",
      string id = ""
    );

    int create_net_heartbeat_v2(
      string dbName,
      string srcNode,
      string dstNode,
      bool   ipv6 = false,
      string viaPath = "",
      string id = ""
    );

    int create_serial_heartbeat( string dbName, string srcNode, string dstNode, string viaPath, string id = "" );
    int create_disc_heartbeat(
      string  dbName,
      string  srcNode,
      string  dstNode,
      string  viaPath,
      int     readBlock,
      int     writeBlock,
      string  svcTag  = "",
      string  id      = "" );

   /*
    * Add Heartbeat Pairs
    * -------------------
    */
    int create_net_heartbeat_pair(
      string dbName,
      string srcNode,
      string dstNode,
      string viaPath = "",
      string id = ""
    );
    int create_net_heartbeat_pair_v2(
      string dbName,
      string srcNode,
      string dstNode,
      bool   ipv6 = false,
      string viaPath = "",
      string id = ""
    );

    int create_serial_heartbeat_pair( string dbName, string srcNode, string dstNode, string viaPath, string id = "" );
    int create_disc_heartbeat_pair(
      string  dbName,
      string  srcNode,
      string  dstNode,
      string  viaPath,
      int     readBlock,
      int     writeBlock,
      string  svcTag  = "",
      string  id      = "" );

   /*
    * Remove Heartbeats
    * -----------------
    */
    int delete_net_heartbeat   ( string dbName, string srcNode, string dstNode, string viaPath = "", string id = "" );
    int delete_net_heartbeat_v2(
      string  dbName,
      string  srcNode,
      string  dstNode,
      string  viaPath  = "",
      string  id       = "",
      int     ipType   = 0 );
    int delete_serial_heartbeat( string dbName, string srcNode, string dstNode, string viaPath, string id = "" );
    int delete_disc_heartbeat(
      string  dbName,
      string  srcNode,
      string  dstNode,
      string  viaPath,
      int     readBlock,
      int     writeBlock,
      string  svcTag,
      string  id = "" );

   /*
    * Remove Heartbeat Pairs
    * ----------------------
    */
    int delete_net_heartbeat_pair   ( string dbName, string srcNode, string dstNode, string viaPath = "", string id = "" );
    int delete_net_heartbeat_pair_v2(
      string  dbName,
      string  srcNode,
      string  dstNode,
      string  viaPath = "",
      string  id      = "",
      int     ipType  = 0 );
    int delete_serial_heartbeat_pair( string dbName, string srcNode, string dstNode, string viaPath, string id = "" );
    int delete_disc_heartbeat_pair(
      string  dbName,
      string  srcNode,
      string  dstNode,
      string  viaPath,
      int     readBlock,
      int     writeBlock,
      string  svcTag  = "",
      string  id      = "" );

   /*
    * List Heartbeats
    * ---------------
    */
    vector<vector<string> > list_net_heartbeats   ( string dbName, string srcNode = "" );
    vector<vector<string> > list_serial_heartbeats( string dbName, string srcNode = "" );
    vector<vector<string> > list_disc_heartbeats  ( string dbName, string srcNode = "" );

   /*
    *  Heartbeat ID List
    *  -----------------
    */
    vector<string>          list_heartbeat_link_ids( string dbName );
    vector<vector<string> > list_linked_heartbeats ( string dbName, string id = "" );

    string get_net_heartbeat_link_id(
      string  dbName,
      string  srcNode,
      string  dstNode,
      string  viaPath = "" );

    string get_serial_heartbeat_link_id(
      string  dbName,
      string  srcNode,
      string  dstNode,
      string  viaPath );

    string get_disc_heartbeat_link_id(
      string  dbName,
      string  srcNode,
      string  dstNode,
      string  viaPath,
      int     readBlock,
      int     writeBlock,
      string  svcTag = "" );

   /*
    * Bulk Remove Heartbeats
    * ----------------------
    */
    int delete_net_heartbeats    ( string dbName, string srcNode = "" );
    int delete_serial_heartbeats ( string dbName, string srcNode = "" );
    int delete_disc_heartbeats   ( string dbName, string srcNode = "" );
    int delete_service_heartbeats( string dbName, string service );
    int delete_linked_heartbeats ( string dbName, string idStr );

   /*
    * Single Node Machine Name
    * ------------------------
    */
    string  read_machine_name   ( string dbName );
    int     create_machine_name ( string dbName, string node );
    int     update_machine_name ( string dbName, string node );
    int     delete_machine_name ( string dbName );
  };

  extern RSFDB_Heartbeats rsfdb_heartbeats;

 /* ==========================================================================
  *
  *   Class     : RSFDB_Services
  *   Instance  : rsfdb_services
  *
  * ==========================================================================
  */

 /**
  * @brief Contains functions to write service values to the selected database.
  */
  class RSFDB_Services
  {
  private:

  public:
   /*
    * Class (de-)constructors
    * -----------------------
    */ 
    RSFDB_Services();
    ~RSFDB_Services();

   /*
    * Service Lists
    * -------------
    */
    vector<string>  get_service_list( string dbName );
    vector<string>  get_server_list ( string dbName, string service );
    vector<string>  get_vip_summary ( string dbName, string service = "" );

    vector<string> get_vips_server_summary(
      string  dbName,
      string  service     = "",
      bool    hidePrimary = false
    );

    vector<string> get_vips_server_summary_v2(
      string  dbName,
      string  service     = "",
      bool    hidePrimary = false,
      bool    useInet     = false
    );

   /*
    * Service Priority
    * ----------------
    */
    string  read_service_priority ( string dbName, string service );
    int     set_service_priority  ( string dbName, string service, int priority );
    int     clear_service_priority( string dbName, string service );

   /*
    * Server Priority
    * ---------------
    */
    string  read_server_priority ( string dbName, string service, string server );
    int     set_server_priority  ( string dbName, string service, string server, int priority );
    int     clear_server_priority( string dbName, string service, string server );

   /*
    * Start Monitor
    * -------------
    */
    string read_server_start_monitor ( string dbName, string service, string server );
    int    set_server_start_monitor  ( string dbName, string service, string server, string value );
    int    clear_server_start_monitor( string dbName, string service, string server );

   /*
    * Service Netmask
    * ---------------
    */
    string  read_service_netmask ( string dbName, string service );
    int     set_service_netmask  ( string dbName, string service, string mask );
    int     clear_service_netmask( string dbName, string service );

   /*
    * Service VIP
    * ----------------
    */
    string          read_service_vip( string dbName, string service );
    vector<string>  get_service_vips( string dbName, string service );

    int get_service_vip_inet_family(
      string dbName,
      string service,
      string vip );

    int add_vip_to_service(
      string  dbName,
      string  service,
      string  vip,
      string  ipdevice,
      string  server,
      string  netmask   = "",
      bool    forceIPv6 = false );

    int set_primary_vip(
      string dbName,
      string service,
      string vip );

    int remove_vip_from_service(
      string dbName,
      string service,
      string vip,
      string server = "" );
    
    int remove_ipdevice_from_service(
      string dbName,
      string service,
      string vip,
      string ipdevice,
      string server = "" );

   /*
    * Service Description
    * -------------------
    */
    string  read_service_desc  ( string dbName, string service );
    int     create_service_desc( string dbName, string service, string desc );
    int     update_service_desc( string dbName, string service, string desc );
    int     delete_service_desc( string dbName, string service );

   /*
    * Service Option
    * --------------
    */
    string  read_service_option  ( string dbName, string service );
    int     create_service_option( string dbName, string service, string option );
    int     update_service_option( string dbName, string service, string option );
    int     delete_service_option( string dbName, string service );

   /*
    * Service Initial Timeout
    * -----------------------
    */
    string  read_service_initimeout  ( string dbName, string service  );
    int     create_service_initimeout( string dbName, string service, int ito );
    int     update_service_initimeout( string dbName, string service, int ito );
    int     delete_service_initimeout( string dbName, string service  );

   /*
    * Service Run Timeout
    * -------------------
    */
    string  read_service_runtimeout  ( string dbName, string service );
    int     create_service_runtimeout( string dbName, string service, int rto );
    int     update_service_runtimeout( string dbName, string service, int rto );
    int     delete_service_runtimeout( string dbName, string service );

   /*
    * Service Mount Points
    * --------------------
    */
    vector<string>  get_service_mount_points     ( string dbName, string service, bool hideLinked = false );
    string          read_service_mount_point_pool( string dbName, string service, string mp );
    int             create_service_mount_point   ( string dbName, string service, string mp );
    int             delete_service_mount_point   ( string dbName, string service, string mp = "" );

   /*
    * Service Environment Variables
    * -----------------------------
    */
    vector<string>  get_setenv_list      ( string dbName, string service );
    string          read_service_setenv  ( string dbName, string service, string key );
    int             create_service_setenv( string dbName, string service, string key, string value );
    int             update_service_setenv( string dbName, string service, string key, string value );
    int             delete_service_setenv( string dbName, string service, string key );

   /*
    * Service GUID
    * ------------
    */
    string  read_service_guid  ( string dbName, string service );
    int     create_service_guid( string dbName, string service, string guid );
    int     update_service_guid( string dbName, string service, string guid );
    int     delete_service_guid( string dbName, string service );

   /*
    * Service Reservation Drives
    * --------------------------
    */
    vector<string>  get_reservation_drives  ( string dbName, string service );
    int             create_reservation_drive( string dbName, string service, string drive );
    int             delete_reservation_drive( string dbName, string service, string drive );

   /*
    * Service Additional Pools
    * ------------------------
    */
    string          read_additional_pool_guid       ( string dbName, string service, string name );
    string          read_additional_pool_mount_point( string dbName, string service, string name );
    vector<string>  get_additional_pools            ( string dbName, string service );

    int create_additional_pool (
      string dbName,
      string service,
      string name,
      string guid = "-",
      string mount = "" );

    int delete_additional_pool( string dbName, string service, string name );

   /*
    * Service IP Device
    * -----------------
    */
    string          read_service_ipdevice  ( string dbName, string service, string node );
    vector<string>  get_extra_ipdevice_list( string dbName, string service, string node );

    vector<string>  get_ipdevice_vips(
      string  dbName,
      string  service,
      string  node,
      string  ipdevice    = "",
      bool    hidePrimary = false );

    vector<vector<string> > get_ipdevice_vips_inet(
      string  dbName,
      string  service,
      string  node,
      string  ipdevice    = "",
      bool    hidePrimary = false );

    string read_ipdevice_for_vip(
      string dbName,
      string service,
      string node,
      string vip );

    string read_additional_vip_netmask(
      string dbName,
      string service,
      string node,
      string ipdevice,
      string vip );

   /*
    * IP Device Logical Operators
    * ---------------------------
    */
    string get_ipdevice_logic(
      string dbName,
      string service,
      string node,
      string interface );

    int set_ipdevice_logic(
      string dbName,
      string service,
      string node,
      string interface,
      string logicOperator );

    int reset_ipdevice_logic(
      string dbName,
      string service,
      string node,
      string interface );

    int bulk_reset_ipdevice_logic(
      string dbName,
      string service = "",
      string node = "",
      string interface = "" );

   /*
    * Custom Monitor Tags
    * -------------------
    */
    string read_monitor_expression ( string dbName, string service, string node );
    string read_custom_monitor_tag ( string dbName, string service, string node );
    int    set_custom_monitor_tag  ( string dbName, string service, string node, string tag );
    int    reset_custom_monitor_tag( string dbName, string service, string node );
  
   /*
    * FC Device Monitor Logic
    * -----------------------
    */
    string read_fcdevice_logic  ( string dbName );
    int    update_fcdevice_logic( string dbName, string logic );
    int    delete_fcdevice_logic( string dbName );

   /*
    * Service Storage Type
    * --------------------
    */
    string read_service_storage_type ( string dbName, string service );
    int    set_service_storage_type  ( string dbName, string service, string type );
    int    reset_service_storage_type( string dbName, string service );

   /*
    * Delete Service
    * --------------
    */ 
    int delete_service( string dbName, string service );

  };

  extern RSFDB_Services rsfdb_services;


/*
** GUI Users Table Columns
** -----------------------
*/
#define GUI_USR_COL   "username"
#define GUI_REAL_COL  "realname"
#define GUI_EML_COL   "email"
#define GUI_PWD_COL   "password"
#define GUI_ROL_COL   "role"
#define GUI_LLI_COL   "lastlogin"
#define GUI_CNT_COL   "logincount"
#define GUI_ENA_COL   "enabled"

  enum gui_user_roles
  {
    USER_ROLE_VIEW = 0,
    USER_ROLE_OPS,
    USER_ROLE_ADMIN
  };

 /**
  * @brief      Class for GUI Users
  */
  class GUI_Users
  {
  private:
    GUI_Users( const GUI_Users& other );      // non construction-copyable
    GUI_Users& operator=( const GUI_Users& ); // non copyable

    int _perform_sql_cmd_gui_users(
      string dbName, string sqlCmd, int * changes, int * sqlRet );

    string _sqlite_read_gui_users_column( string dbName, string username, string column );

    string  _read_user_info    ( string dbName, string username, string key );
    int     _read_user_info_int( string dbName, string username, string key );

    int _update_user_info(
      string dbName, string username, string key, string value );

    int _is_last_admin( string dbName, string username );

    vector<string> _list_admins(
      string dbName, string username = "", int *result = NULL );

  public:
    GUI_Users();

   /*
    * User Account
    * ------------
    */
    vector<string> list_users( string dbName );

    int create_user(
      string  dbName,
      string  username,
      string  realname  = "",
      string  email     = "",
      string  password  = "",
      int     role      = 0,
      bool    enabled   = true );

    int delete_user( string dbName, string username );

   /*
    * Realname
    * --------
    */
    string  read_realname   ( string dbName, string username );
    int     update_realname ( string dbName, string username, string name );

   /*
    * E-mail Address
    * --------------
    */
    string  read_email_address  ( string dbName, string username );
    int     update_email_address( string dbName, string username, string address );
    int     delete_email_address( string dbName, string username );

   /*
    * Password Hash
    * -------------
    */
    string  read_password_hash  ( string dbName, string username );
    int     update_password_hash( string dbName, string username, string password );

   /*
    * User role
    * ---------
    */ 
    string  get_role_string ( int id );
    int     read_role_id    ( string dbName, string username );
    int     update_role     ( string dbName, string username, int role );

   /*
    * User login
    * ----------
    */
    int login_user      ( string dbName, string username );
    int read_last_login ( string dbName, string username );
    int read_login_count( string dbName, string username );

   /*
    * Account Enable/Disable
    * ----------------------
    */ 
    bool  is_account_enabled( string dbName, string username );
    int   disable_account   ( string dbName, string username );
    int   enable_account    ( string dbName, string username );

  };

 /*
  * ============================================================================
  * ||                    _____  ____     __    _  __                         ||
  * ||                   / ___/ / __ \   / /   (_)/ /_ ____                   ||
  * ||                   \__ \ / / / /  / /   _  / ,_// , /                   ||
  * ||                  ___/ // /_/ /_ / /__ / // /_ / __/                    ||
  * ||                 /___ / \____,_//____//_//___//___/                     ||
  * ||                                                                        ||
  * ============================================================================
  */

 /**
  * @enum         sql_table_syntax
  *
  * @brief        Contains refrences for the SQLite tables supported by the database
  */
  enum sql_table_syntax
  {
    TBL_VIPS = 0,
    TBL_GLOBAL,
    TBL_GUIUSERS,
    TBL_HEARTBEATS,
    TBL_MOUNTPOINTS,
    TBL_PROPS,
    TBL_SERVERS,
    TBL_SERVICES,
    TBL_SETENV,
    TBL_COUNT
  };

 /**
  * @enum          setenv_type
  *
  * @brief         Contains refrences for the supported environment variable types
  */
  enum setenv_type
  {
    ENV_STANDARD,
    ENV_ADDIT_POOL,
    ENV_ZPOOL_GUID,
    ENV_RESERVATION
  };

 /**
  * @enum          cli_hbt_action
  *
  * @brief         Contains refrences for the supported heartbeat function actions
  */
  enum cli_hbt_action
  {
    HBT_WRITE = 0,
    HBT_REMOVE
  };

 /**
  * @brief      Contains valid flag values for sql_literal() function.
  * 
  *               NOCASE  - Perform case insensitive comparison
  *               INVERSE - Check the key does NOT contain the value
  *               BARE    - Use bare value (i.e Do not add quotes)
  */
  struct SQLiteValue
  {
    enum Flags{
      NOCASE    = (1<<0),
      INVERSE   = (1<<1),
      BARE      = (1<<2)
    };
  };

 /* ==========================================================================
  *
  *   Class     : SQLiteDB
  *   Instance  : rsfsql
  *
  * ==========================================================================
  */

  typedef std::vector< std::vector< std::vector <std::string> > > rsfSQLVec;

   /**
    * Provides functions that act as a bridge between rsfcdb and SQLite. The
    * functions within this class are all private and are only accessible by
    * friend classes. The functions that should be used by the friend functions
    * are listed under the RECOMMENDED FUNCTIONS section. The functions listed
    * under the INTERNAL FUNCTIONS section should only be used within this class.
    * Their use within friend classes should be avoided but may be used sparingly
    * and as a last resort. If a function within a friend class needs to use
    * these functions excessively, they should either be placed in this class
    * or a new wrapper should be written in the RECOMMENDED section of this
    * class.
    * 
    * The functions that directly perform SQL commands on the database return a
    * vector of vectors of vectors of strings. The structure of this object is:
    * @verbatim
    * (
    *   (
    *     <vector> row_1_vector:
    *     (
    *       <vector> column_1_vector : ( <string> column_1_name, <string> column_1_value ),
    *       <vector> column_2_vector : ( <string> column_2_name, <string> column_2_value ),
    *       <vector> column_n_vector : ( <string> column_n_name, <string> column_n_value )
    *     ),
    *     
    *     ...
    *     
    *     <vector> row_n_vector:
    *     (
    *       <vector> column_1_vector : ( <string> column_1_name, <string> column_1_value ),
    *       <vector> column_2_vector : ( <string> column_2_name, <string> column_2_value ),
    *       <vector> column_n_vector : ( <string> column_n_name, <string> column_n_value )
    *     ) 
    *   )
    * )
    * @endverbatim
    * 
    * This object is only accessed using funtions within this class. It is not
    * reccommended to access this object manually; you should only access this
    * object via the rsfSQLVec hanfdler functions provided. To suppress the
    * typename, this object is given a typedef of rsfSQLVec ( i.e RSF SQLite
    * vector type ).
    * @brief  Provides SQLite convienence functions for database operations.
    */


  class SQLiteDB
  {
   /* Friend declarations to allow the public exposed classes to access the
    * private SQLite functionality they require to function correctly.
    */
    friend class SQLite_Global;
    friend class SQLite_Heartbeat;
    friend class SQLite_Services;
    friend class CDB_CLI;
    friend class GUI_Users;
    friend class RSFDB_Prop;
    friend class RSFDB_Config;

  private:

     /* ========================================================================
      * INTERNAL FUNCTIONS
      * ========================================================================
      */
      SQLiteDB( const SQLiteDB& other );        // non construction-copyable
      SQLiteDB& operator=( const SQLiteDB& );   // non copyable

     /* 
      * Table - Enum to String
      * ----------------------
      */
      string  get_table_syntax  ( sql_table_syntax s );

     /*
      * Open/Close Database
      * -------------------
      */
      int       journal_handler ( string dbName );
      sqlite3 * sqlite_open     ( string dbName , bool readOnly   );
      int       sqlite_close    ( sqlite3 * db  , int vanilla = 0 );

     /*
      * Validate database
      * -----------------
      */ 
      int  is_valid_db( string dbName );

     /*
      * Database Snapshots
      * ------------------
      */
      string _get_snapshot_name( string dbName );

      int _process_snapshot ( string dbSrc, string dbDst, int action );
      int create_snapshot   ( string dbName );
      int restore_snapshot  ( string dbName, bool cleanup = false );
      int delete_snapshot   ( string dbName );

      int create_snapshot_v2   ( string dbName, string snapshotName );
      int restore_snapshot_v2  ( string dbName, string snapshotName, bool cleanup = false );
      int delete_snapshot_v2   ( string snapshotName );

     /*
      * Database Operations
      * -------------------
      */
      static int  lock_handler  ( void* v, int count );

      rsfSQLVec db_exec(
        sqlite3 * db,
        string sql_str,
        int * result,
        bool rawMode = true
      );

      int  disable_foreign_key_support ( sqlite3 * db );
      int  enable_foreign_key_support  ( sqlite3 * db );
      bool has_foreign_key_support     ( sqlite3 * db );

      vector<string> foreign_key_check ( sqlite3 * db );

      int         sqlite_init   ( sqlite3 * db );
      int         unset_vanilla ( sqlite3 * db );

     /*
      * Table Operations
      * ----------------
      */
      int drop_sqlite_table    ( string dbName, sql_table_syntax tbl );
      int rebuild_sqlite_table ( sqlite3 * db, sql_table_syntax tbl );
      int rename_sqlite_table  ( sqlite3 * db, string oldName, string newName );

     /*
      * Table Check
      * -----------
      */
      int check_table        ( sqlite3 * db, sql_table_syntax table, bool isTemp = false );
      int check_table_exists ( sqlite3 * db, sql_table_syntax table );

      vector<string> get_common_table_columns( sqlite3 * db, string tbl1, string tbl2 );

     /* ========================================================================
      * RECOMMENDED FUNCTIONS
      * ========================================================================
      */

     /* 
      * Table - Enum to String
      * ----------------------
      */
      string  get_table_name  ( sql_table_syntax s, bool isTemp = false );

     /*
      * rsfSQLVec Handler Functions
      * ---------------------------
      */
      string          get_first_value ( rsfSQLVec vvv, bool lowerCase = false, string prefix = "" );
      vector<string>  get_all_values  ( rsfSQLVec vvv, bool lowerCase = false, string prefix = "" );

     /*
      * SQLite String Processing
      * ------------------------
      */ 
      string sql_escape_quotes( string s );

     /*
      * SQLite Literal Construtor
      * -------------------------
      */
      string sql_literal( string key, string value, int flags = 0 );

     /*
      * SQLite Command Functions
      * ------------------------
      */
      int rsfcdb_return_status_handler( int ret, int * sqlRet );

      bool has_column_error( sqlite3 * db, int result );

      int perform_sql_cmd(
        string            dbName,
        sql_table_syntax  table,
        string            sqlCmd,
        int               *changeCount,
        int               vanilla = 0 );

     /*
      * SQLite Query Functions
      * ----------------------
      */

      rsfSQLVec exec_sql_query(
        string  dbName,
        string  sqlCmd,
        int     *result = NULL,
        int     *cdbRes = NULL );

      rsfSQLVec perform_sql_select(
        string            dbName,
        string            selectStr,
        sql_table_syntax  table,
        string            whereStr  = "",
        string            additStr  = "",
        int               *result   = NULL,
        int               *cdbRes   = NULL );

      int table_exists(
        string dbName,
        string table );

     /*
      * List Functions
      * --------------
      */
      vector<string> list_tables( string dbName, bool showAll = false );
      
     /*
      * Database Reset
      * --------------
      */
      int reset_config_db  ( string dbName );
      int reset_rsf_cluster( string dbName );

      int _rename_private_heartbeats(
        string                     dbName,
        string                     node,
        vector< vector< string > > privHbts = vector< vector <string> >() );

      int _rename_table_node(
        string            dbName,
        sql_table_syntax  table,
        string            colName,
        string            oldNode,
        string            newNode );

      int _delete_table_node(
        string            dbName,
        sql_table_syntax  table,
        string            colName,
        string            node );


    public:
      int rename_cluster_node(
        string                      dbName,
        string                      oldNode,
        string                      newNode,
        vector< vector< string > >  privHbts = vector< vector <string> >() );

      int delete_cluster_node(string dbName, string oldNode);

    public:
      SQLiteDB();
    };

    extern SQLiteDB rsfsql;

   /* ==========================================================================
    *
    *   Class     : SQLite_Global
    *   Instance  : sqlite_gbl
    *
    * ==========================================================================
    */

   /**
    * @brief Contains functions to configure global values in an SQLite3 database
    */
    class SQLite_Global
    {
    private: 
     /*
      * SQLite Command Handler
      * ----------------------
      */
      int perform_sql_cmd_global(
        string  dbName,
        string  key,
        string  value,
        bool    update = false );

     /*
      * SQLite Query Handler
      * --------------------
      */
      string sqlite_read_global(
        string  dbName,
        string  key );

     /*
      * SQLite Write Key Handler
      * ------------------------
      */
      int sqlite_write_global_str(
        string  dbName,
        string  key,
        string  value );

      int sqlite_write_global_int(
        string  dbName,
        string  key,
        int     value );

     /*
      * SQLite Update Key Handler
      * -------------------------
      */
      int sqlite_update_global_str(
        string  dbName,
        string  key,
        string  value );

      int sqlite_update_global_int(
        string  dbName,
        string  key,
        int     value );

     /*
      * SQLite Delete Key Handler
      * -------------------------
      */
      int sqlite_delete_global(
        string dbName,
        string key );

    public:
      SQLite_Global();
      ~SQLite_Global();

     /*
      *  Cluster Name
      *  ------------
      */
      string  read_cluster_name  ( string dbName );
      int     create_cluster_name( string dbName, string name );
      int     update_cluster_name( string dbName, string name );
      int     delete_cluster_name( string dbName );

     /*
      *  Cluster Description
      *  -------------------
      */
      string  read_cluster_description  ( string dbName );
      int     create_cluster_description( string dbName, string desc );
      int     update_cluster_description( string dbName, string desc );
      int     delete_cluster_description( string dbName );

     /*
      *  Cluster Identifier
      *  ------------------
      */
      string  read_cluster_identifier  ( string dbName );
      int     create_cluster_identifier( string dbName, string ident );
      int     update_cluster_identifier( string dbName, string ident );
      int     delete_cluster_identifier( string dbName );

     /*
      *  GLOBAL Heartbeat Backoff
      *  ------------------------
      */
      string  read_disc_hb_backoff  ( string dbName );
      int     create_disc_hb_backoff( string dbName, int start, int multiplier, int max );
      int     update_disc_hb_backoff( string dbName, int start, int multiplier, int max );
      int     delete_disc_hb_backoff( string dbName );

     /*
      *  Cluster Poll Time
      *  -----------------
      */
      string  read_poll_time  ( string dbName );
      int     create_poll_time( string dbName, int pt );
      int     update_poll_time( string dbName, int pt );
      int     delete_poll_time( string dbName );

     /*
      *  Cluster Real Time
      *  -----------------
      */
      string  read_realtime  ( string dbName );
      int     create_realtime( string dbName, int rt );
      int     update_realtime( string dbName, int rt );
      int     delete_realtime( string dbName );

     /*
      *  Cluster Event Notify
      *  --------------------
      */
      string  read_event_notify  ( string dbName );
      int     create_event_notify( string dbName, string en );
      int     update_event_notify( string dbName, string en );
      int     delete_event_notify( string dbName );

     /*
      *  Cluster IP Device Monitor
      *  -------------------------
      */
      string  read_ipdevice_monitor  ( string dbName );
      int     create_ipdevice_monitor( string dbName, int pu, int pd );
      int     update_ipdevice_monitor( string dbName, int pu, int pd );
      int     delete_ipdevice_monitor( string dbName );

     /*
      *  Cluster FC Device Monitor
      *  -------------------------
      */
      string  read_fcdevice_monitor  ( string dbName );
      int     create_fcdevice_monitor( string dbName, int pu, int pd );
      int     update_fcdevice_monitor( string dbName, int pu, int pd );
      int     delete_fcdevice_monitor( string dbName );

     /*
      *  Heartbeat Timeout
      *  -----------------
      */
      string  read_heartbeat_timeout  ( string dbName );
      int     create_heartbeat_timeout( string dbName, int ht );
      int     update_heartbeat_timeout( string dbName, int ht );
      int     delete_heartbeat_timeout( string dbName );
    };

    extern SQLite_Global sqlite_gbl;


   /* ==========================================================================
    *
    *   Class     : SQLite_Heartbeat
    *   Instance  : sqlite_hbt
    *
    * ==========================================================================
    */

    /**
     * @brief Contains functions to configure heartbeat values in an SQLite3 database
     */
    class SQLite_Heartbeat
    {
     /* Friend declaration to allow the public exposed class to access SQLite
      * heartbeat functionality.
      */ 
      friend class RSFDB_Heartbeats;
      
     /* Friend declaration to allow the private SQLite Service class access to
      * private functions. This is required for the remove services command to
      * function correctly.
      */ 
      friend class SQLite_Services;

    private:
     /*
      * Heartbeat Type enum to String
      * -----------------------------
      */ 
      string get_type_str ( heartbeat_type ht );

     /*
      * Pair Path String
      * ----------------
      */ 
      string get_pair_via_str ( string viaStr );

     /*
      * CLI Handler Function
      * --------------------
      */ 
      int cli_hbt_action_v2(
        heartbeat_type  hbtType,
        int             action,
        bool            pair,
        string          dbName,
        string          id,
        string          srcNode,
        string          dstNode,
        string          viaPath     = "",
        int             readBlock   = 0,
        int             writeBlock  = 0,
        string          svcTag      = "" );

     /*
      * SQLite Command Handler
      * ----------------------
      */
      int perform_sql_cmd_heartbeat(
        string  dbName,
        string  sqlCmd,
        int     *changes,
        int     *sqlRet = NULL );

     /*
      * SQLite Hearbeat Commands
      * ------------------------
      */
      int write_heartbeat(
        string          dbName,
        heartbeat_type  type,
        string          srcNode,
        string          dstNode,
        string          viaStr  = "",
        string          tagStr  = "",
        string          idStr   = "",
        bool            force   = false );

      int delete_heartbeat(
        string          dbName,
        heartbeat_type  type,
        string          srcNode = "",
        string          dstNode = "",
        string          viaStr  = "",
        string          tagStr  = "",
        string          idStr   = "" );

      int _delete_net_heartbeat(
        string          dbName,
        string          srcNode,
        string          dstNode,
        string          viaStr,
        string          idStr,
        bool            pair,
        heartbeat_type  type = HBT_NET );

     /*
      * SQLite Heartbeat Query
      * ----------------------
      */ 
      vector<string> get_heartbeat_config_lines(
        string          dbName,
        string          src,
        heartbeat_type  type,
        string          path = "" );

     /*
      * Linked heartbeat Query
      * ----------------------
      */
      string _get_heartbeat_link_id(
        string          dbName,
        heartbeat_type  type,
        string          srcNode,
        string          dstNode,
        string          viaPath,
        string          svcTag );

     /*
      * Network heartbeat list
      * ----------------------
      */
      vector<vector<string> > _list_net_heartbeats(
        string dbName,
        string src,
        bool showFamily );

    public:

     /*
      * Class (De-)Constructors
      * -----------------------
      */ 
      SQLite_Heartbeat();
      ~SQLite_Heartbeat();

     /*
      *  List Heartbeat Source Nodes
      *  ---------------------------
      */
      vector<string> get_heartbeat_nodes( string dbName, bool format_node );

     /*
      *  List Heartbeat Config Lines
      *  ---------------------------
      */
      vector<string> get_net_config_lines   ( string dbName, string src );
      vector<string> get_serial_config_lines( string dbName, string src );
      vector<string> get_disc_config_lines  ( string dbName, string src );

     /*
      *  List Heartbeat Elements
      *  -----------------------
      */
      vector<vector<string> > list_net_heartbeats   ( string dbName, string src );
      vector<vector<string> > list_net_heartbeats_v2( string dbName, string src );
      vector<vector<string> > list_serial_heartbeats( string dbName, string src );
      vector<vector<string> > list_disc_heartbeats  ( string dbName, string src );

     /*
      *  Heartbeat ID List
      *  -----------------
      */
      vector<string>          list_heartbeat_link_ids( string dbName );
      vector<vector<string> > list_linked_heartbeats( string dbName, string id );

      string get_net_heartbeat_link_id(
        string dbName, string srcNode, string dstNode, string viaPath );

      string get_serial_heartbeat_link_id(
        string dbName, string srcNode, string dstNode, string viaPath );

      string get_disc_heartbeat_link_id(
        string  dbName,
        string  srcNode,
        string  dstNode,
        string  viaPath,
        int     readBlock,
        int     writeBlock,
        string  svcTag );

     /*
      *  Create Individual/paired Heartbeats
      *  -----------------------------------
      */
      int create_net_heartbeat(
        string  dbName,
        string  srcNode,
        string  dstNode,
        string  viaStr,
        string  idStr,
        bool    pair,
        bool    force,
        bool    ipv6 );

      int create_serial_heartbeat(
        string  dbName,
        string  srcNode,
        string  dstNode,
        string  viaStr,
        string  idStr,
        bool    pair,
        bool    force );

      int create_disc_heartbeat(
        string  dbName,
        string  srcNode,
        string  dstNode,
        string  viaStr,
        int     readBlock,
        int     writeBlock,
        string  tagStr,
        string  idStr,
        bool    pair,
        bool    force );

     /*
      *  Remove Individual/paired Heartbeats
      *  -----------------------------------
      */
      int delete_net_heartbeat(
        string  dbName,
        string  srcNode,
        string  dstNode,
        string  viaStr,
        string  idStr,
        bool    pair );

      int delete_serial_heartbeat(
        string  dbName,
        string  srcNode,
        string  dstNode,
        string  viaStr,
        string  idStr,
        bool    pair );

      int delete_disc_heartbeat(
        string  dbName,
        string  srcNode,
        string  dstNode,
        string  viaStr,
        int     readBlock,
        int     writeBlock,
        string  tagStr,
        string  idStr,
        bool    pair );

      int delete_linked_heartbeats( string dbName, string idStr );

     /*
      *  Bulk Remove Heartbeats
      *  ----------------------
      */
      int bulk_delete_net_heartbeats   ( string dbName, string srcNode = "" );
      int bulk_delete_serial_heartbeats( string dbName, string srcNode = "" );
      int bulk_delete_disc_heartbeats  ( string dbName, string srcNode = "" );
      int bulk_delete_tagged_heartbeats( string dbName, string service = "" );

     /*
      *  Single Node Machine Name
      *  ------------------------
      */
      string  read_machine_name  ( string dbName );
      int     create_machine_name( string dbName, string node );
      int     update_machine_name( string dbName, string node );
      int     delete_machine_name( string dbName );
    };

    extern SQLite_Heartbeat sqlite_hbt;


   /* ==========================================================================
    *
    *   Class     : SQLite_Services
    *   Instance  : sqlite_svc
    *
    * ==========================================================================
    */

   /**
    * @brief Contains functions to configure service values in an SQLite3 database
    */
    class SQLite_Services
    {
     /* Friend declaration to allow the public exposed class and the CLI to
      * access private SQLite service functionality.
      */ 
      friend class RSFDB_Services;
      friend class CDB_CLI;
      friend class SQLiteDB;
    private:

     /* ========================================================================
      *
      * Command Handlers
      * ----------------
      *
      * These are commands that call the SQLite class to execute the commands.
      * The results are then "handled" by performing any additional actions
      * required due to the results obtained.
      */

     /*
      * SQLite Servers Command Handler
      * ------------------------------
      */ 
      int _perform_sql_cmd_servers(
        string  dbName,
        string  sqlCmd,
        int     *changes  = NULL,
        int     *sqlRet   = NULL );

     /*
      * SQLite Services Command Handlers
      * --------------------------------
      */ 
      int _handle_sql_cmd_services(
        string  dbName,
        string  sqlCmd,
        string  service,
        string  key,
        bool    delMode );

      int _perform_sql_cmd_services(
        string  dbName,
        string  sqlCmd,
        int     *changes  = NULL,
        int     *sqlRet   = NULL );
      
      string _read_service_param(
        string  dbName,
        string  service,
        string  key,
        bool    show_err = true );

      int _write_service_param(
        string  dbName,
        string  service,
        string  key,
        string  value,
        bool    force = false );

      int _remove_service_param(
        string  dbName,
        string  service,
        string  key,
        bool    force = false );

      int _handle_service_param_remove( string dbName, string service );
      int _add_service_row            ( string dbName, string service );

      int _delete_service             ( string dbName, string service );

     /*
      * SQLite VIPs Command Handler
      * ---------------------------
      */ 
      int _perform_sql_cmd_vips(
        string  dbName,
        string  sqlCmd,
        int     *changes  = NULL,
        int     *sqlRet   = NULL );

      int _create_vip(
        string  dbName,
        string  service,
        string  vip,
        string  ipdevice,
        string  netmask,
        string  server,
        int     serverPriority,
        bool    ipv6Addr,
        bool    update );

      int _delete_vip(
        string  dbName,
        string  service,
        string  vip,
        string  server,
        string  ipdevice,
        bool    netmask );

      vector<string> _list_service_vips(
        string  dbName,
        string  service,
        string  ipdevice,
        string  server,
        bool    hidePrimary );

      vector<vector<string> > _list_service_vips_inet(
        string  dbName,
        string  service,
        string  ipdevice,
        string  server,
        bool    hidePrimary );

     /*
      * Primary Status Handlers
      * -----------------------
      */
      string _read_primary_param(
        string  dbName,
        string  service,
        string  key,
        bool    quiet = false );

      int _handle_remove_vip_primary_status(
        string  dbName,
        string  service,
        string  vip,
        string  server );

      int _process_primary_status(
        string  dbName,
        string  service,
        string  vip );

     /*
      * SQLite Environment Variables Command Handlers
      * ---------------------------------------------
      */ 
      int _handle_service_env_param(
        string  dbName,
        string  sqlCmd,
        string  service,
        bool    quiet = false );

      int _perform_sql_cmd_setenv(
        string  dbName,
        string  sqlCmd,
        int     *changes  = NULL,
        int     *sqlRet   = NULL );

     /*
      * SQLite Mount Point Command Handlers
      * -----------------------------------
      */ 
      vector<string>  _get_mountpoint_param_list(
        string  dbName,
        string  service,
        string  pool,
        bool    show_err );

      string _read_mountpoint_param(
        string dbName,
        string service,
        string mount,
        string key,
        bool showErr );

      int _handle_sql_cmd_mountpoints(
        string  dbName,
        string  sqlCmd,
        string  service,
        bool    delMode );

      int _perform_sql_cmd_mountpoints(
        string  dbName,
        string  sqlCmd,
        int     *changes  = NULL,
        int     *sqlRet   = NULL );

      int _create_mountpoint_param(
        string  dbName,
        string  service,
        string  mount,
        string  pool = "",
        bool    force = false );

      int _delete_mountpoint_param(
        string  dbName,
        string  service,
        string  mount = "",
        string  pool = "" );

     /*
      * SQLite Servers Command Handler
      * ------------------------------
      */ 
      bool _check_service_node_exists(
        string  dbName,
        string  service,
        string  node    = "",
        bool    showErr = true );

     /*
      * Environment Variable Handlers
      * -----------------------------
      */
      string _read_service_env_param(
        string    dbName,
        string    service,
        string    key,
        unsigned  envType,
        bool      showErr );

      int _create_service_env_param(
        string      dbName,
        setenv_type type,
        string      name,
        string      key,
        string      value,
        bool        force );

      int _delete_service_env_param(
        string  dbName,
        string  service,
        string  key );

      string _get_setenv_type_str( setenv_type t );

     /*
      * IP Device Command Handlers
      * --------------------------
      */
      string _get_service_ipdevice_param(
        string  dbName,
        string  key,
        string  service,
        string  node,
        string  ipdevice,
        string  vip = "" );

      vector<string>  _get_service_ipdevices_list(
        string dbName,
        string service,
        string node,
        primary_vip_filter primary );

     /*
      * Priority Handlers
      * -----------------
      */ 
      bool   _process_service_priority       ( int priority );
      string _read_server_priority_param     ( string dbName, string service, string server );
      int    _create_server_priority_param   ( string dbName, string service, string server, int pri );
      int    _remove_server_priority_param   ( string dbName, string service, string server );

      string          _parse_custom_monitor_value       ( string service, string server, string tag );
      vector<string>  _get_validated_monitor_tokens     ( string tag );
      string          _generate_custom_monitor_tag_value( string tag );

      string _read_server_start_monitor_param  ( string dbName, string service, string server );
      int    _create_server_start_monitor_param( string dbName, string service, string server, string value );
      int    _remove_server_start_monitor_param( string dbName, string service, string server );

     /*
      * Value Range Checker Functions
      * -----------------------------
      */
      int _check_initimeout_range ( int ini );
      int _check_runtimeout_range ( int run );

     /* ========================================================================
      *
      * Front-end Class Functions
      * -------------------------
      *
      * These are the functions that should be called to perform actions with
      * services in SQLite. These are also the functions that should be called
      * from other classes within RSFCDB.
      */

     /*
      * 'Global' Service
      * ----------------
      */
      vector<string> get_service_list       ( string dbName );
      vector<string> get_servers_for_service( string dbName, string service );
      vector<string> get_services_for_server( string dbName, string server );

      int delete_service( string dbName, string service );

     /*
      * Vips & Nodes
      * ------------
      */
      vector<string> get_service_vips(
        string  dbName,
        string  service,
        string  server      = "",
        bool    hidePrimary = false );

      vector<string> get_vip_summary(
        string  dbName,
        string  service     = "",
        bool    hidePrimary = false );

      vector<string> get_vips_server_summary(
        string  dbName,
        string  service,
        bool    hidePrimary = false );

      vector<string> get_vips_server_summary_v2(
        string  dbName,
        string  service,
        bool    hidePrimary = false,
        bool    useInet     = false
      );

      int get_service_vip_inet_family(
        string  dbName,
        string  service,
        string  vip );

      int write_service_vip(
        string  dbName,
        string  service,
        string  vip,
        string  ipdevice,
        string  server          = "",
        string  netmask         = "",
        int     serverPriority  = 0,
        bool    ipv6Addr        = false );

      int update_service_vip(
        string  dbName,
        string  service,
        string  vip,
        string  ipdevice,
        string  server,
        string  netmask         = "",
        int     serverPriority  = 0,
        bool    ipv6Addr        = false );

      int delete_service_vip(
        string  dbName,
        string  service,
        string  vip,
        string  server = "" );

     /*
      * Primary VIP Convenience Functions
      * ---------------------------------
      */
      vector<string> get_primary_vip_servers( string dbName, string service, bool lowerCase = false );
      string         read_primary_vip       ( string dbName, string service, bool quiet = false );
      string         read_primary_netmask   ( string dbName, string service, bool quiet = false );
      int            set_primary_vip        ( string dbName, string service, string vip );

     /*
      *  Service Vip Netmasks
      *  --------------------
      */
      string  get_service_vip_netmask   ( string dbName, string service, string vip, string server );
      int     set_primary_netmask       ( string dbName, string service, string mask, bool update = false );
      int     delete_service_vip_netmask( string dbName, string service, string vip, string server = "" );

     /*
      *  Service Description
      *  -------------------
      */
      string  read_service_description   ( string dbName, string service );
      int     write_service_description  ( string dbName, string service, string desc );
      int     update_service_description ( string dbName, string service, string desc );
      int     remove_service_description ( string dbName, string service );

     /*
      *  Service Vip Option Line
      *  -----------------------
      */
      string  read_service_option  ( string dbName, string service );
      int     write_service_option ( string dbName, string service, string optStr );
      int     update_service_option( string dbName, string service, string optStr );
      int     remove_service_option( string dbName, string service );

     /*
      *  Service Initial Timeout
      *  -----------------------
      */
      string  read_service_initimeout  ( string dbName, string service );
      int     write_service_initimeout ( string dbName, string service, int ini );
      int     update_service_initimeout( string dbName, string service, int ini );
      int     remove_service_initimeout( string dbName, string service );

     /*
      *  Service Running Timeout
      *  -----------------------
      */
      string  read_service_runtimeout  ( string dbName, string service );
      int     write_service_runtimeout ( string dbName, string service, int run );
      int     update_service_runtimeout( string dbName, string service, int run );
      int     remove_service_runtimeout( string dbName, string service );

     /*
      *  Service Mount Points
      *  --------------------
      */
      vector<string> list_service_mount_points    ( string dbName, string service, string pool = "", bool hideLinked = false );
      string         read_service_mount_point_pool( string dbName, string service, string mp );
      int            write_service_mount_point    ( string dbName, string service, string mp, string pool = "" );
      int            update_service_mount_point   ( string dbName, string service, string mp, string pool = "" );
      int            remove_service_mount_point   ( string dbName, string service, string mp = "", string pool = "" );

     /*
      *  Service Environment Variables
      *  -----------------------------  
      */
      vector<string> get_setenvs_of_type(
        string      dbName,
        string      service,
        setenv_type type,
        bool        listValues = false );
      
      string  read_service_setenv  ( string dbName, string service, string key );
      int     write_service_setenv ( string dbName, string service, string key, string value );
      int     update_service_setenv( string dbName, string service, string key, string value );
      int     remove_service_setenv( string dbName, string service, string key );

     /*
      *  Service Zpool Guid
      *  ------------------
      */
      string  read_service_guid  ( string dbName, string service );
      int     write_service_guid ( string dbName, string service, string guid );
      int     update_service_guid( string dbName, string service, string guid );
      int     remove_service_guid( string dbName, string service );

     /*
      *  Additional Zpool Guid
      *  ---------------------
      */
      vector<string>  get_additional_pools            ( string dbName, string service );
      string          read_additional_pool_guid       ( string dbName, string service, string name );
      string          read_additional_pool_mount_point( string dbName, string service, string name );

      int write_additional_pool(
        string dbName,
        string service,
        string name,
        string guid,
        string mount );  

      int remove_additional_pool( string dbName, string service, string name ); 

     /*
      *  Reservation Drive
      *  -----------------
      */  
      vector<string>  get_reservation_drives  ( string dbName, string service );
      int             write_reservation_drive ( string dbName, string service, string name );
      int             remove_reservation_drive( string dbName, string service, string name );

     /*
      * IP Device
      * ---------
      */
      vector<string> get_service_ipdevices( string dbName, string service, string node, bool primary );

      vector<string> get_service_ipdevices_v2(
        string dbName,
        string service,
        string node,
        primary_vip_filter primary = VIP_ALL );

      vector<string>  get_ipdevice_vips(
        string  dbName,
        string  service,
        string  ipdevice,
        string  server      = "",
        bool    hidePrimary = false );

      vector<vector<string> > get_ipdevice_vips_inet(
        string  dbName,
        string  service,
        string  ipdevice,
        string  server      = "",
        bool    hidePrimary = false );

      string read_service_ipdevice(
        string  dbName,
        string  service,
        string  node );

      string read_service_ipdevice_netmask(
        string  dbName,
        string  service,
        string  node,
        string  ipdevice,
        string  vip );

      string read_service_ipdevice_vip(
        string  dbName,
        string  service,
        string  node,
        string  ipdevice );

      string read_service_ipdevice_for_vip(
        string dbName,
        string service,
        string node,
        string vip );

      int delete_service_ipdevice(
        string  dbName,
        string  service,
        string  vip,
        string  ipdevice,
        string  server = "" );

     /*
      * Service Priority
      * ----------------
      */
      int     check_priority_range   ( int pri );

      string  read_service_priority  ( string dbName, string service );
      int     set_service_priority   ( string dbName, string service, int pri );
      int     remove_service_priority( string dbName, string service );

     /*
      * Server Priority
      * ---------------
      */
      string  read_server_priority  ( string dbName, string service, string server );
      int     set_server_priority   ( string dbName, string service, string server, int pri );
      int     remove_server_priority( string dbName, string service, string server );

     /*
      * Start Monitor
      * -------------
      */
      string  read_server_start_monitor  ( string dbName, string service, string server );
      int     set_server_start_monitor   ( string dbName, string service, string server, string value );
      int     clear_server_start_monitor( string dbName, string service, string server );

     /*
      * IP Device Monitor Logic
      * -----------------------
      */
      bool _is_valid_logical_operator( string logic );

      int _create_interface_logic_param(
        string dbName,
        string service,
        string server,
        string interface,
        string logicOperator,
        bool   bulkMode );

      string get_ipdevice_logic  ( string dbName, string service, string server, string interface );
      int    set_ipdevice_logic  ( string dbName, string service, string server, string interface, string logicOperator );
      int    reset_ipdevice_logic( string dbName, string service, string server, string interface );

      int bulk_reset_ipdevice_logic (
        string dbName,
        string service,
        string server,
        string interface );

     /*
      * Custom Start Monitor
      * --------------------
      */
      string read_monitor_expression( string dbName, string service, string server );
      string read_custom_monitor_tag( string dbName, string service, string server );

      int set_custom_monitor_tag(
        string dbName,
        string service,
        string server,
        string tag );

      int reset_custom_monitor_tag( string dbName, string service, string server );

     /*
      * FC Device Monitor Logic
      * -----------------------
      */
      string read_fcdevice_logic  ( string dbName );
      int    update_fcdevice_logic( string dbName, string logic );
      int    delete_fcdevice_logic( string dbName );

     /*
      * Service CLI Printing Funtions
      * -----------------------------
      */
      int list_server_priorities_cli ( string dbName, string service );
      int list_service_priorities_cli( string dbName );

     /*
      * Service Storage Type
      * --------------------
      */
      string read_service_storage_type ( string dbName, string service );
      int    set_service_storage_type  ( string dbName, string service, string type );
      int    reset_service_storage_type( string dbName, string service );

    public:
      SQLite_Services();
      ~SQLite_Services();
    };
    extern SQLite_Services sqlite_svc;

   /* ==========================================================================
    *
    *   Struct      : CLICmd
    *   Description : Structure used to construct a reference to a CLI command.
    *                 This allows all information and help for the CLI commands
    *                 to be stored together, making it easier to maintain the
    *                 source code.
    *   Usage       : CLICmd( cmd_name, cmd_args, cmd_help );
    *   Args        :
    *    cmd_name   string          Command name as called from the CLI
    *    cmd args   vector<string>  Valid usage argument syntax for command.
    *    cmd_help   string          Brief explaination of the command's function
    *
    * ==========================================================================
    */

   /**
    * @brief Structure used to construct a reference to a CLI command
    */
    struct CLICmd
    {
    public:
     /**
      * @brief The CLI command name 
      */
      string cmdName;

     /**
      * @brief String vector containing the accepted command syntax.
      */
      vector<string> cmdArgs;

     /**
      * @brief String containing a brief explaination of the command's function.
      */
      string cmdHelp;

     /**
      * @brief Constructor for an empty object.
      */
      CLICmd( ) : cmdName(""), cmdArgs(vector<string>()), cmdHelp("") {}

     /**
      * @brief Constructor that assigns the internal strucure variables.
      */
      CLICmd( const string a, vector<string> b, const string c ) : cmdName(a), cmdArgs(b), cmdHelp(c) {}
    };

    enum cmd_action
    {
      CMD_WRITE = 0,
      CMD_UPDATE,
      CMD_DELETE
    };

   /* ==========================================================================
    *
    *   Class     : CDB_CLI
    *   Instance  : cli
    *
    * ==========================================================================
    */

   /**
    * @brief Provides functions that allows interaction with the API via the
    *        command line.
    */
    class CDB_CLI
    {
     /* Friend declaration to allow the public exposed class to access the CLI
      * class. This allows the classes to store CLI specific actions in the CLI
      * class itself.
      */ 
      friend class RSFDB_Prop;
      friend class RSFDB_Global;
      friend class RSFDB_Heartbeats;
      friend class RSFDB_Services;
      friend class RSFDB_Config;
      friend class Debug;

    private:
      CDB_CLI( const CDB_CLI& other );        // non construction-copyable
      CDB_CLI& operator=( const CDB_CLI& );   // non copyable

     /*
      * Internal CLI Variables
      * ----------------------
      */ 
      int   dbSelect;         // Database selection value
      
      string dbPathStr        ; // Database path
     /*
      * CLI Flag Indicators
      * -------------------
      */ 
      bool dbSelectFlag;    // -d flag ( database selection ) provided
      bool showDescFlag;    // -I flag was called (for prop descriptions)
      bool dbPathFlag;      // -t flag was called

      bool  silentFlag;     // Silent mode enabled

      int   verboseFlag;    // Verbose output
      int   pairFlag;       // Pair mode for heartbeats
      int   allFlag;        // Pair mode for heartbeats
      int   batchFlag;      // Batch mode for prop values
      int   inet4Flag;      // Remove heartbeats with an IPv4 address
      int   inet6Flag;      // Treat heartbeat/vip as IPv6 address
      int   usePrimary;     // Respect primary vips in generated config file

      bool  singleNodeFlag; // Single node cluster mode
      bool  dbLockFlag;     // Boolean that indicates if db locking should be used
      bool  monTagsFlag;    // Generate monitor tags in generated config file
      bool  timeStampFlag;  // Show timestamps for database properties
      bool  envListFlag;    // Print env export commands from list-props

     /*
      * CLI Flag Arguments
      * ------------------ 
      */
      string referenceOpt ; //  Reference string (for -r flag)
      string defWriteOpt  ; //  Default write value (for -D flag)
      string settingsOpt  ; //  Settings file path
      string netmaskOpt   ; //  Netmask for new vip
      string interfaceOpt ; //  Network interface for the command
      string nodeOpt      ; //  Node/server name for the command
      string hbtDevicePath; //  Hearbeat device path
      string hbtTagOpt;     //  Service tag for disc heartbeat
      string hbtIdOpt;      //  Heartbeat link ID
      
      vector<string> nodeOptList; // List of nodes for the command

      std::map<int, char*> longArgNames;
      std::map<int, char*> longArgReqNames;

    /* -------------------------------------------------------------------------
     *
     * CLI Class Functions
     * ===================
     */   

     /*
      * Unsupported Database Handler
      * ----------------------------
      */
      int handle_unsupported_db ( string trace );

     /*
      * CLI Help/Usage Info Functions
      * -----------------------------
      */
      CLICmd  get_cmd_obj       ( int i );
      int     cmd_hidden        ( int cmd );
      void    get_cmd_help      ( CLICmd cmd );
      void    get_cmd_syntax    ( CLICmd cmd );
      void    print_switch_help ( int s );
      string  get_alt_cmd       ( string s );
      bool    valid_cmd_str     ( const string input, const string cmd );
      int     get_cmd_id        ( string str );
      int     print_cmd         ( string str );
      void    print_cmd_list    ( );

     /*
      * RSFCDB Status/Health Functions
      * ------------------------------
      */
      int   set_db_type     ( string dbType );
      void  check_db_status ( );

     /*
      * Settings file Functions
      * -----------------------
      */ 
      int write_settings_file ( string s_path = CDB_SETTINGS_FILE );
      int read_settings_file  ( string s_path = CDB_SETTINGS_FILE );

     /*
      * Custom Property Funtions
      * ------------------------
      */ 
      int list_custom_props_cli( string dbName , bool print_ts );
      int list_props_env_cli   ( string dbName, bool hideInvalid = true );

      int write_prop_cli(
        string  dbPath,
        string  key,
        string  value,
        string  desc,
        string  ref,
        bool    update );

      int write_prop_desc_cli(
        string  dbPath,
        string  key,
        string  desc );

      int prop_batch_cmd_cli(
        string        dbPath,
        vector<char*> args,
        string        refStr,
        int           action );

     /*
      * Heartbeat Functions
      * -------------------
      */
      vector<string> get_hbt_read_list(
        string  dbName,
        int     hbType,
        string  srcNode = "" );

      int print_formatted_hb_list(
        string  db,
        string  type,
        string  src = "" );

      int print_linked_hb_list(
        string  db,
        string  id );

      int read_hbts_cli(
        string  dbPath,
        int     hbType,
        string  srcNode = "" );

      int net_heartbeat_bulk_cmd(
        string          dbPath,
        bool            delMode,
        vector<string>  nodes,
        string          devicePath = "",
        string          id         = "" );

      int serial_heartbeat_bulk_cmd(
        string          dbPath,
        bool            delMode,
        vector<string>  nodes,
        string          devicePath,
        string          id = "" );

      int disc_heartbeat_bulk_cmd(
        string          dbPath,
        bool            delMode,
        vector<string>  nodes,
        string          devicePath,
        int             readBlock,
        int             writeBlock,
        string          serviceTag,
        string          id = "" );

      int extract_disc_device_blocks(
        string  rawPath,
        string  *path,
        int     *readBlk,
        int     *writeBlk );

     /*
      * Service/Server Functions
      * ------------------------
      */
      int print_configured_vips_cli( string dbName, string service = "" );

     /*
      * CLI Priority Functions
      * ----------------------
      */ 
      int print_service_priorities( string dbName );
      int print_server_priorities ( string dbName, string service );

     /*
      * CLI Flag Processing
      * -------------------
      */
      vector<int>   argint       ( string str );
      vector<char*> filter_argv  ( char **argv , int argc );
      vector<char*> get_args_list( int ac , char* av[] , string optStr );
      int           check_opt    ( int ac , char* av[] );

     /*
      * Type Hnadler Functions
      * ----------------------
      */ 
      int string_handler( string s );
      int vector_handler( vector<string> v );
      int cmd_handler   ( string str );
      
    public:
      CDB_CLI();

      int getSelectedDB();

      int cmd_line ( int argc, char* argv[] );
    };

    extern CDB_CLI cli;


   /* ==========================================================================
    *
    *   Class     : Debug
    *   Instance  : cdblog
    *
    * ==========================================================================
    */

   /**
    * This class is non-copyable and should be accessed using the global cdblog
    * object.
    * @brief Handle's API debugging and writing to the log file. 
    */
    class Debug
    {
     /* Friend class declarations to allow the public classes and the private
      * SQLite functins to access the debugging class.
      */
      friend class CDB_CLI;
      friend class RSFDB_Config;
      friend class RSFDB_Global;
      friend class RSFDB_Heartbeats;
      friend class RSFDB_Services;
      friend class RSFDB_Prop;

     /* Friend declarations to allow all the SQLite classes full access to the
      * debugging class.
      */ 
      friend class SQLiteDB;
      friend class SQLite_Global;
      friend class SQLite_Heartbeat;
      friend class SQLite_Services;      
      friend class GUI_Users;
    private:
     /*
      * Non-copyable Class Constructors
      * -------------------------------
      */
      Debug( const Debug& other );        // non construction-copyable
      Debug& operator=( const Debug& );   // non copyable
      
     /*
      * Debugging Output Stream
      * -----------------------
      */ 
      std::ofstream dout;

     /*
      * Debugging Flags
      * ---------------
      */ 
      bool  debug_mode;
      bool  debug_config_write;   // Write debug info to config file
      bool  debug_high;           // High level of debug file logging

     /*
      * API Global Error Code
      * ---------------------
      */ 
      int cdb_error;

     /*
      * Variables for log break characters
      * ----------------------------------
      */
      string logBreak;
      string logSubBreak;

     /*
      * Initial class check function
      * ----------------------------
      */
      void _init_chks();

     /*
      * Public Flag Triggers
      * --------------------
      */
      string  get_log_break     ();
      string  get_log_subbreak  ();

     /*
      * Error setting functions
      * -----------------------
      */ 
      int  set_err ( int errRef );

    public:
     /*
      * Public Class Constructor
      * ------------------------
      */  
      Debug();

     /*
      * Public Debug class status flags
      * -------------------------------
      */ 
      static bool firstWrite;
      static bool cliMode;
      static bool useColour;

     /*
      * Public Debug class error flags
      * ------------------------------
      */
      static bool printErr;
      static bool logErr;

      void write    ( std::ostream& s, bool highLvl );
      void error    ( std::string trace , std::ostream& s );
      void verbose  ( cdb_verbose_levels verboseLevel, std::ostream& s );

     /*
      * Public Flag Triggers
      * --------------------
      */
      bool  trigger_debug_mode    ( int status , int high_debug = 0 );
      bool  trigger_error_print   ( int status );
      bool  trigger_error_logging ( int status );

     /*
      * Public Debugging Status Read Functions
      * --------------------------------------
      */
      int get_verbose_status        ();
      int get_error_print_status    ();
      int get_error_logging_status  ();
      int get_config_debug_status   ();
      int get_err                   ();
      
    };
    extern Debug cdblog;

   /* ==========================================================================
    *
    * Template Functions
    * ------------------
    *
    *   Any template functions are declared and implemented here.
    *
    * ==========================================================================
    */

   /**
    *  Converts a string array to a vector
    *  @param  arr String array to be converted.
    *  @return String vector.
    */
    template <size_t sze>
    vector<string> arrToVec( const string (&arr)[sze])
    {
      vector<string> v;
      v.assign( arr, arr+sze );
      return v;
    }

   /* ==========================================================================
    *
    * Custom Functions
    * ----------------
    *
    *   Custom function declarations.
    *
    * ==========================================================================
    */

   /*
    * Vector Functions
    * ----------------
    */
    bool            vec_has_val     ( string value, vector<string> sv );
    void            show_vec        ( vector<string> vec, string pre = "" );
    vector<string>  tokenizer       ( string str, char del );
    string          stringify_vector( vector<string> v, string sep = "," );
   /*
    * String Functions
    * ----------------
    */
    bool        check_param               ( string str );
    string      filter_str                ( string str, char del );
    bool        found_at                  ( string str, size_t pos, char chr );
    vector<int> get_disc_hb_backoff_ints  ( string s );
    string      get_disc_hb_backoff_str   ( int start, int multiplier, int max );
    vector<int> get_monitor_ints          ( string s );
    string      get_monitor_str           ( int pu, int pd );
    int         get_longest_str_size      ( vector<string> v) ;
    bool        is_valid_int              ( string str, bool allow_negative = true );
    string      lowerStr                  ( string s );
    string      upperStr                  ( string s );
    bool        is_alnumscore_string      ( const std::string &str );
    string      replace_substr            ( string str, string findStr, string substr );
    string      substr_word               ( string str, size_t pos );
  
   /*
    * System Functions
    * ----------------
    */
    int print_pid_process( int pid = getpid() );
    
   /*
    * Time & Date Functions
    * ---------------------
    */ 
    string  timestr( time_t tin = 0 );

   /*
    * Conversion Functions
    * --------------------
    */
    string    inttostring ( int val );
    int       strtoint    ( string str, int *res = NULL );
    long int  strtolint   ( string str, int *res = NULL );

   /*
    * File Functions
    * --------------
    */
    bool  file_exists   ( const std::string& name );
    int   get_path_type ( const string path       );
    int   touch_file    ( const string filename   );

   /*
    * Networking Functions
    * --------------------
    */
    string  process_addr( string addrStr, bool machine_mode = false, bool single_process = false );

    string  strip_addr              ( string addrStr, bool machine_mode = false);
    int     validate_ipaddr_version ( const string ipaddr, int *version );

  }

  extern "C"{

  }
#endif /* LIBRSFCDB_H_ */
