#!/bin/bash
#
# Note! *** Run this script on NODEA ***

# This array defines some pre-configured commands that can be run after a failover
# These will be displayed in a menu, unless the user disables interactive mode.
# The user can specify a custom command to run, or choose from this list of options.
#
# To add more commands, just add items to this array in the format "<test name>:<unix command>"
# If the string contains more than one ':', the first will be used as a separator.
PRE_CONFIGURED_TEST[1]="Big Data test:/opt/HAC/bin/zfs_stress.sh <POOL>"
PRE_CONFIGURED_TEST[2]='RSF Log message:/opt/HAC/RSF-1/bin/rsfcli msg info Completed failover test ${CNT}'

OPTION_LIST=(NODEA NODEB POOL LOGDIR LUCNT SLEEPTIME NUMFAILS MOVE COMMAND)

# Help texts to be printed when the user is prompted to enter something
NODEA_HELP="name of the node this script is running on"
NODEB_HELP="name of the other node in the cluster"
POOL_HELP="comma separated list of services to failover"
LOGDIR_HELP="a suitable directory to store log files"
LUCNT_HELP="number of LUs in volume - type 0 if there are none"
SLEEPTIME_HELP="amount of time in seconds to wait between failovers"
NUMFAILS_HELP="number of failovers to perform before exiting (0 = infinite)"
MOVE_HELP="rsfcli command use move instead of quickmove (y or n)"
COMMAND_HELP="command to run after each failover (type ';' for no command)"
COMMAND_HELP="command to run after each failover \
             \n\t0) No command \
             \n\t1) Pre-configured test \
             \n\t2) Custom command \
             \nPlease select an option or enter for default"

show_usage(){
	cat <<-END >&2
	Usage: $(basename $0) [options]
	    -a <name>    - Node A name (run this script on node A!)
	    -b <name>    - Node B name
	    -c <num>     - Lun count - should match the number of luns you have on volume
	    -l <dir>     - Suitable log directory (should exist on both machines)
	    -p <name>    - Pool name, or comma separated list of pools
	    -s <time>    - Amount of time to wait between failovers (seconds)
	    -n <num>     - Number of failovers to perform (0 for infinite)
	    -m           - Use rsfcli move instead of quickmove (faster failover)
	    -x <command> - Specify a command to run after each failover (';' for nothing)
	    -d           - Use default options without prompting
	    -r           - Suppress bold output
	    -h           - Show this message and stop
	END
}

# Function to test one of the options. It takes two arguments: the name and the value
# of the argument. It will return 0 if the argument is OK, 1 for not OK, and 2 for an error.
# This function can change the option value (e.g. in the case of COMMAND where it can use a 
# separate sub menu). 
test_option(){
	retval=1
	if [ $# -ne 2 -o "$2" = "" ]; then
	    return $retval
	fi
	OPTION_NAME="$1"
	OPTION="$2"
	case $OPTION_NAME in
	    NODEA) retval=0;;
	    NODEB) retval=0;;
	    POOL) retval=0;;
	    LOGDIR)
                if [ -d "$OPTION" ]; then
	            retval=0
	        fi 2>/dev/null;;
	    LUCNT|SLEEPTIME|NUMFAILS)
                if [ $OPTION -eq $OPTION ]; then
	            retval=0
                fi 2>/dev/null;;
            MOVE)
                if [ "$OPTION" = "Y" -o "$OPTION" = "y" ]; then
                    OPTION=move
                elif [ "$OPTION" = "N" -o "$OPTION" = "n" ]; then
                    OPTION=quickmove
                fi

                if [ "$OPTION" = "move" -o  "$OPTION" = "quickmove" ]; then
                    retval=0
                fi;;
            COMMAND)
                # If option is a number, then it can't be a command and must be from
                # the previous menu. Decide what to do with it here.
                if [ $OPTION -eq $OPTION ]; then
                    case $OPTION in
                        0) OPTION=";" ;;
                        1)
                           echo "Pre-configured commands:"
                           echo -e "\t0) Cancel"
                           for (( i=1; i <= ${#PRE_CONFIGURED_TEST[@]}; i++ )); do
                               echo -e "\t$i) ${PRE_CONFIGURED_TEST[$i]%%:*}"
                           done
                           input=
                           while [ -z "$input" ] || ! [ $input -le ${#PRE_CONFIGURED_TEST[@]} ]; do
                               echo -n "Enter an option: "
                               read input
                           done 2>/dev/null
                           if [ $input -eq 0 ]; then
                               return $retval
                           elif [ $input -le ${#PRE_CONFIGURED_TEST[@]} ]; then
                               OPTION="${PRE_CONFIGURED_TEST[$input]#*:}"
                           fi ;;
                        2) echo -n "Enter custom command: "
                           read -e OPTION 2>&1;;
                    esac
                fi 2>/dev/null
                # Now we have a COMMAND option from either:
                # - The processing above
                # - The original option, which if not a number, will have avoided the above.
                #   - Most likely, this will have been either from the -x command line option,
                #     or from a default option stored in a file.
                # We now need to test the option to see if it can be run.
                base_command="${OPTION%%\ *}"                    # Strip off any command line arguments
                if [ "${base_command}" = ";" ]; then             # Special case for a ';' - no command will be run
                    retval=0
                elif which ${base_command} >/dev/null 2>&1; then # The command may not start with a path -
                    retval=0                                     # can it be found in the PATH directories
                elif [ -x ${base_command} ] && ssh root@${NODEB} "[ -x ${base_command} ]"; then
                    retval=0                                     # Finally, see if the command is an executable file
                else
                    echo "Command not found: $OPTION"
                fi 2>/dev/null;;
	esac

        # If the command has not been found to be runnable, don't save it.
        # When this function returns false, the user will be asked to enter something else
        if [ $retval -eq 0 ]; then
            export ${OPTION_NAME}="${OPTION}"
        fi

	return $retval
}

# Process options
OPTION_SAVE_FILE=/opt/HAC/RSF-1/etc/.failtest.options # Try to load options from a file
if [ -f ${OPTION_SAVE_FILE} ]; then
    echo "Loading defaults..."
    source ${OPTION_SAVE_FILE}
fi

OPTION_CHANGED=0
while getopts "a:b:p:c:s:n:x:l:drh" name # Command line arguments to change options
do
    case $name in 
        a) NODEA=${OPTARG}     && OPTION_CHANGED=1 ;;
        b) NODEB=${OPTARG}     && OPTION_CHANGED=1 ;;
        p) POOL=${OPTARG}      && OPTION_CHANGED=1 ;;
        l) LOGDIR=${OPTARG}    && OPTION_CHANGED=1 ;;
        c) LUCNT=${OPTARG}     && OPTION_CHANGED=1 ;;
        s) SLEEPTIME=${OPTARG} && OPTION_CHANGED=1 ;;
        n) NUMFAILS=${OPTARG}  && OPTION_CHANGED=1 ;;
        m) MOVE=${OPTARG}      && OPTION_CHANGED=1 ;;
        x) COMMAND="${OPTARG}" && OPTION_CHANGED=1 ;;
        d) USE_DEFAULTS=true  ;;
        r) SUPPRESS_BOLD=true ;;
        h|?) show_usage
           exit               ;;
    esac
done

# Check all options have been set - if any haven't prompt the user to enter them
for option in ${OPTION_LIST[@]}
do
    help_text_var="${option}_HELP"
    # Don't prompt the user if the option has been set AND they have used the -d option.
    if [ "${!option}" == "" -o "${USE_DEFAULTS}" != "true" ]; then
        # The first test_option will always fail because it is given nothing for the option value
        # This means the user will always be prompted at least once.
        new_option=
        while ! test_option "${option}" "${new_option}"; do
            echo -en "${option} (${!help_text_var}) [${!option}]: " | sed "s/ \[\]//g"
            read -e new_option >&2
            if [ "${new_option}" != "" -a "${new_option}" != "${!option}" ]; then
                OPTION_CHANGED=1
            fi
            # If the user has not entered anything, it could be that they want the default
            new_option="${new_option:-${!option}}"
            # Now go back and test the revised 'new_option'
        done
        # test_option has returned true, so it will have set the relavent variable
    fi
    # Move on to the next variable
done

# All options must be set by now, write them to file to be used next time
if [ $OPTION_CHANGED -eq 1 ]; then
    read -p "Save new options to file? [y]" save
    save=${save:-"Y"}
    if [ $(echo ${save} | tr 'y' 'Y') == Y ]; then
	cat <<-EOF | tee ${OPTION_SAVE_FILE} | sed "s/=/:\t/g"
	#### Default options for $(basename $0) ####
	NODEA=${NODEA}
	NODEB=${NODEB}
	POOL=${POOL}
	LOGDIR=${LOGDIR}
	LUCNT=${LUCNT}
	SLEEPTIME=${SLEEPTIME}
	NUMFAILS=${NUMFAILS}
	MOVE=${MOVE}
	COMMAND='${COMMAND}'
	#########################################
	EOF
    fi
fi

#
# NO USER EDITABLE PARTS BELOW
# 
PATH=/opt/HAC/RSF-1/bin:$PATH
export PATH

echo "HAC HA-Plugin fail over test script."
echo "First     node: ${NODEA}"
echo "Second    node: ${NODEB}"
echo "Fail-over pool: ${POOL}"
echo "Fail-over time: ${SLEEPTIME}"

LOGA=${LOGDIR}/${NODEA}
LOGB=${LOGDIR}/${NODEB}
CNT=0

mkdir -p ${LOGA}
mkdir -p ${LOGB}

RSFCLI=/opt/HAC/RSF-1/bin/rsfcli

POOL="`echo ${POOL} | tr ',' ' '`"

bold()
{
    start=
    end=
    if [ "${SUPPRESS_BOLD}" != "true" ]; then
        start=$(tput bold)
        end=$(tput sgr0)
    fi

    echo ${start}$*${end}
}

nodes_to_auto() 
{
    for pool in ${POOL}; do
	${RSFCLI} -i=0 -h ${NODEA} auto ${pool}
	${RSFCLI} -i=0 -h ${NODEB} auto ${pool}
    done
    sleep 2
}

save_logs()
{
	LUC=${1}
	local NODE=${2}

        echo "Luns on ${NODE} = ${LUC}"

        if [ "${LUCNT}" != "${LUC}" ] ; then
		echo "**** LU COUNT WRONG"
	fi
	printf "Collecting logs from %s..." ${NODEA}
        (cd /opt/HAC/RSF-1/log/; tar cf - `ls -1 | egrep 'stmfha'` | bzip2 -9 > ${LOGA}/logs-`date +%d-%M-%Y_%H:%M:%S`.tar.bz2)
	echo "done"
	printf "Collecting logs from %s..." ${NODEB}
	ssh ${NODEB} "(cd /opt/HAC/RSF-1/log/; tar cf - \`ls -1 | egrep 'stmfha'\` | bzip2 -9)" > ${LOGB}/logs-`date +%d-%M-%Y_%H:%M:%S`.tar.bz2
	echo "done"

	nodes_to_auto
}

service_running()
{
    if [ $# -lt 2 ]; then
        echo "service_running called with incorrect args" >&2
        return 2
    fi

    NODE=$1
    shift
    pools=$*

    for pool in $pools; do
        if ! ${RSFCLI} modes|grep "Service ${pool}.*running.*${NODE}" > /dev/null 2>&1; then
            return 1
        fi
    done

    return 0
}

do_failover()
{
    TONODE=${1}
    [ "${TONODE}" = "${NODEA}" ] && FROMNODE=${NODEB} || FROMNODE=${NODEA}

    COUNT=${CNT}
    RSFLOG=/opt/HAC/RSF-1/log/rsfmon.log

    # Perform failover
    echo "Moving to ${TONODE}, test ${COUNT}"
    for pool in ${POOL}; do
        ${RSFCLI} -i=0 ${MOVE} ${pool} ${TONODE} >/dev/null 2>&1
    done

    # Wait until failover has completed
    TOTAL_FAILOVER_TIME=0
    until service_running ${TONODE} ${POOL}; do
        if [ $(( ${TOTAL_FAILOVER_TIME} % 10 )) -eq 0 ]; then
            echo "Failover in progress..."
        fi
        sleep 1
        # Count number of seconds for full failover
        TOTAL_FAILOVER_TIME=$(( ${TOTAL_FAILOVER_TIME} + 1 ))
    done

    # If possible, use tac to search through the log in reverse (quicker for big log files)
    if which tac >/dev/null; then
        TIME_CMD="tac ${RSFLOG} | grep 'Total run time' | head -n 1"
    else
        TIME_CMD="grep 'Total run time' ${RSFLOG} | tail -n 1"
    fi

    # Get service start and stop times
    LOCAL_TIME="`eval ${TIME_CMD}`"      # Either start or stop on local node
    REMOTE_TIME="`ssh root@${NODEB} \""${TIME_CMD}"\"`" # Either start or stop on remote node
    if [ ${TONODE} = ${NODEA} ]; then
        STOP_TIME="${REMOTE_TIME##*]}"
        START_TIME="${LOCAL_TIME##*]}"
    else
        STOP_TIME="${LOCAL_TIME##*]}"
        START_TIME="${REMOTE_TIME##*]}"
    fi

    echo ${STOP_TIME}
    echo ${START_TIME}
    # Total failover time is NOT equal to stop time + start time - there is a gap between stop and start
    bold "Total failover time: ${TOTAL_FAILOVER_TIME} seconds"
}

# This function runs a user defined command on the machine specified.
# It takes a single argument - the machine on which to run the command.
# Output from this function will be saved to a separate log file.
do_extra()
{
    if [ $# -ne 1 ]; then
        echo "do_extra called with incorrect number of arguments!" >&2
        return 1
    elif [ "$1" != "$NODEA" -a "$1" != "$NODEB" ]; then
        echo "do_extra called with unrecognised node name" >&2
        return 1
    fi

    local NODE=${1}
    EXTRALOG=${LOGDIR}/${NODE}/post-failover-${COUNT}.log

    SED_COMMAND=
    for opt in ${OPTION_LIST[@]}; do
        [ "$opt" = "COMMAND" ] && continue;
        SED_COMMAND="${SED_COMMAND} -e s%<${opt}>%`echo ${!opt} | awk '{print $1}'`%g"
    done

    COMMAND="$(echo ${COMMAND}|sed ${SED_COMMAND})"

    # Special case for ';', user doesn't want to do anything
    if [ "${COMMAND}" = ";" ]; then
        return
    fi

    # Print start time to console and log
    eval echo "`date`: Running \"${COMMAND}\" on ${NODE}" | tee ${EXTRALOG}

    # Run command and print output to console and log
    if [ "${NODE}" = "${NODEA}" ]; then
        eval ${COMMAND} 2>&1               # Local machine
    elif [ "${NODE}" = "${NODEB}" ]; then
        ssh root@${NODE} "`eval echo ${COMMAND}` 2>&1" # Remote machine
    fi | tee -a ${EXTRALOG}

    # Print end time to console and log
    echo "`date`: Done" | tee -a ${EXTRALOG}
}

do_sleeptime()
{
    echo "Waiting ${SLEEPTIME} seconds"
    sleep ${SLEEPTIME}
}

while [ ${CNT} -lt ${NUMFAILS} -o ${NUMFAILS} -eq 0 ] ; do
    echo
    CNT=`expr ${CNT} + 1`
    if [ $(( $CNT % 2 )) -eq 1 ]; then
        NODE=${NODEB}
        LU_CMD="ssh ${NODE} 'stmfadm list-lu | wc -l'"
    else
        NODE=${NODEA}
        LU_CMD="stmfadm list-lu | wc -l"
    fi
    do_failover ${NODE} ${CNT}
    LUS=`eval $LU_CMD`
    LUS=${LUS:-"-1"}
    echo "`date`: ${NODE} LUs = ${LUS}" >> ${LOGDIR}/lucount.txt
    save_logs ${LUS} ${NODE}
    do_extra ${NODE}
    do_sleeptime
done

