#!/bin/sh

. /opt/HAC/bin/rsf.sh

DISK_LIST=/opt/HAC/RSF-1/bin/disk_list
MAX_CONCURRENT_TKOWNS=3

if [ ! -z "${PROP_SCSI2_RETRY_COUNT}" -a "${PROP_SCSI2_RETRY_COUNT}" != "0" ]; then
	SCSI2_RETRIES="-R ${PROP_SCSI2_RETRY_COUNT}"
else
	SCSI2_RETRIES=
fi

check_res_disabled()
{
	#
	# This variable tells us which exit code to use if reservations are
	# disabled for any reason
	#
	CODE=$1
	#
	# If this is a VMware cluster, we don't use reservations, so just exit OK
	#
	if [ "${State}" != "forcestop" ] && is_vmware; then
		rc_exit ${service} ${CODE} "vCentre: configuration detected - reservations are disabled"
	fi

	#
	# Check if MHDC reservation has been disabled...
	OPT_ZFS_MHDC_DISABLE=${opt_zfs_mhdc_disable:-"n"}
	if match "${OPT_ZFS_MHDC_DISABLE}X" [Yy]X; then
		rc_exit ${service} ${CODE} "*** SCSI reservations disabled for this volume"
	fi

	#
	# We can't use reservations for S3 pools. If all pools in the service are
	# S3 backed, then we don't have anything to do here
	#
	if all_pools_are_s3; then
		rc_exit ${service} ${CODE} "All pools in this service are S3 backed - nothing to do"
	fi

	#
	# If the service name is Cn_NONE and there are no additional pools, then
	# this service controls no zpools, so we don't need to reserve anything
	#
	if [ -z "${ADDITIONAL_ZPOOL_1}" ] && expr "${service}" : 'C[0-9][0-9]*_NONE' > /dev/null; then
		rc_exit ${service} ${RSF_OK} "No pools associated with this service so no reservations needed. Nothing to do"
	fi
}

abort_unfenced_if_prop_true()
{
	svc=$1
	shift
	code=$1
	shift
	msg="$*"
	
	if [ "${PROP_ABORT_UNFENCED_IMPORT}" = "${BOOL_STR_FALSE}" ] ; then
		dated_echo "${msg} - continuing because PROP_ABORT_UNFENCED_IMPORT = ${PROP_ABORT_UNFENCED_IMPORT}" WARN
	else
		notify_watchers ${LOG_CRIT} "RSF_SERVICE_STARTUP" "service=${svc} resource=RES_DRIVES status=FAILED"
		rc_exit ${svc} ${code} ${msg}
	fi
}

# This function is here to limit the number of concurrent TKOWN operations
# although it could be used to limit concurrent instances of anything.
# The first argument is the number of allowed concurrent instances, then
# any arguments after that are the command to run.
#
# Note that the successful file lock is detected when with_lock exits
# with any code that is not 255. Be careful not to use this function to
# run any command that returns 255!
instance_limit()
{
	WITH_LOCK=${PRODUCT_BIN}/with_lock
	
	MAX_PROCESSES=$1
	shift
	command_name=`basename $1`
	lock_prefix="/tmp/${command_name}-lock-"
	
	ret=255
	
	while [ $ret -eq 255 ]; do
		for i in `seq $MAX_PROCESSES`; do
			lockfile=${lock_prefix}${i}
			${WITH_LOCK} -t 0 ${lockfile} $@
			ret=$?
			if [ $ret -eq 255 ]; then
				dated_echo "Failed to lock file ${lockfile}"
			else
				dated_echo "${command_name} completed using lock file ${lockfile}"
				break
			fi
		done
		[ $ret -eq 255 ] && sleep 1
	done
	
	return $ret
}

#
# Given a variable base name, a function and a list of arguments, we work
# through the list of variables calling the function once for each var in
# turn.  The variable base name has the integers appended to it, e.g. the
# base name "N" would  scan N1, N2, N3, etc. The scan stops at the  first
# variable  which is not set.  Variables may be local shell variables  or
# environment variables. The function is called with the supplied (fixed)
# arguments, followed by the value of the variable.
#
# Usage: process_variable_list variable_base function [args...]
process_variable_list()
{
	varbase="$1"
	shift

	varnumber=0;
	while true; do
		varnumber=`expr ${varnumber} + 1`
		varname="${varbase}${varnumber}"
		eval value=\$${varname}
		match x"${value}" x && break
		"$@" "${value}"
	done
}

#
# Check for reservation drives to use for persistent reservation,
# First cehck the file /opt/HAC/RSF-1/etc/.res_drives.<service> and
# if that does not exists or is empty, then check the rsf-1 configuration
# file for the environment variables RSF_RESERVATION_DRIVE_<num>, otherwise
# there is no drives in use, i.e. persistent reservation is not configured.
#
# For MetroHA, only the res_drives mechanism is supported and we have to get
# 2 lists of drives - local and remote.
check_for_reservation_drives()
{
	UNKNOWN_ARG="-u"
	DEVID_ARG=
	DUMMY_DEVID=
	service=$1
	RES_DRIVES_FILE_TO_CHECK="${MHDC_RESERVATION_DRVS}.${service}"
	shift
	while [ $# -gt 0 ]; do
		case "$1" in
			--with-devid)
				DEVID_ARG="-d"
				DUMMY_DEVID=":unknown"
				;;
			--res-file=*)
				RES_DRIVES_FILE_TO_CHECK=`echo $1|cut -d "=" -f "2"`
				;;
			--with-unknowns)
				UNKNOWN_ARG=
				;;
			*)
				# Something went wrong
				rc_exit ${service} ${RSF_ABORT} "Invalid argument to chec_for_reservation_drives"
				;;
		esac
		shift
	done

	if [ "${PROP_MHDC_USE_PGR3}" = "${BOOL_STR_TRUE}" ]; then
		min_drive_count=${PROP_PGR3_DRIVE_COUNT}
	else
		min_drive_count=${PROP_SCSI2_DRIVE_COUNT}
	fi
	res_drives=
	local_res_drives=
	remote_res_drives=

	if is_metroha; then
		dated_echo "MetroHA cluster detected"
		# Get list of local res_drives
		dated_echo "Reading local reservation drive list from ${RES_DRIVES_FILE_TO_CHECK}..."
		local_res_drives="`${DISK_LIST} -f ${RES_DRIVES_FILE_TO_CHECK} -n ${min_drive_count} ${UNKNOWN_ARG} ${DEVID_ARG} -L | tr '\n' ' '`"
		dated_echo "Local reservation drives: ${local_res_drives}"
		# Get list of remote res_drives
		dated_echo "Reading remote reservation drive list from ${RES_DRIVES_FILE_TO_CHECK}..."
		remote_res_drives="`${DISK_LIST} -f ${RES_DRIVES_FILE_TO_CHECK} -n ${min_drive_count} ${UNKNOWN_ARG} ${DEVID_ARG} -R | tr '\n' ' '`"
		dated_echo "Remote reservation drives: ${remote_res_drives}"
	else
		if [ -s "${RES_DRIVES_FILE_TO_CHECK}" ] ; then
			dated_echo "Reading reservation drive list from ${RES_DRIVES_FILE_TO_CHECK}..."
			res_drives="`${DISK_LIST} -f ${RES_DRIVES_FILE_TO_CHECK} -n ${min_drive_count} ${UNKNOWN_ARG} ${DEVID_ARG} | tr '\n' ' '`"
		else
			dated_echo "Reading reservation drive list from RSF-1 environment variables..."
			res_drives_raw="`process_variable_list RSF_RESERVATION_DRIVE_ echo`"
			for drive in ${res_drives_raw}; do
				if expr "${drive}" : ^c[0-9][0-9]*t..*d[0-9][0-9]*$ >/dev/null; then
					drive=/dev/rdsk/${drive}s0
				fi
				res_drives="${res_drives}${drive}${DUMMY_DEVID} "
			done
		fi
		dated_echo "Reservation drives: ${res_drives}"
	fi
}

reserve_drives()
{
	service=$1
	shift
	drives=$@
	drive_cnt=$#

	METRO_ARG=
	if is_metroha; then
		METRO_ARG="-M"
	fi

	
	awk_scr='/\/dev\/rdsk\/|\/devices\//{
			gsub("^.*ioctl\\(", "");
			sub(", ", " ", $0);
			if ($NF == "OK")
				print $1;
		}'
	
	if [ $drive_cnt -eq 0 ]; then
		dated_echo "Nothing to reserve!" ERROR
		return 1
	fi
	
	dated_echo "About to reserve drives: ${drives}" 
	
	#
	# Decide on fail fast depending upon config setting
	if [ ! -z "${PROP_MHDC_USE_FAILFAST}" -a "${PROP_MHDC_USE_FAILFAST}" = "${BOOL_STR_FALSE}" ] ; then
		dated_echo "Not applying failfast due to property setting of PROP_MHDC_USE_FAILFAST: ${PROP_MHDC_USE_FAILFAST}"
		ARG_FAILFAST=""
	else
		dated_echo "Applying failfast due to property setting of PROP_MHDC_USE_FAILFAST: ${PROP_MHDC_USE_FAILFAST}"
		ARG_FAILFAST="ENFAILFAST ${ENFAILFAST_MICRO}"
	fi
	
	if [ "${PROP_MHDC_USE_PGR3}" = "${BOOL_STR_TRUE}" ]; then
		# PGR3 reservations
		dated_echo "PROP_MHDC_USE_PGR3 database property in use - testing device(s) for PGR3 support"
		if ${PROG_MHDC} -d ${drives} -c PGR3_TEST; then
			dated_echo "OK to set PGR3 reservations on the following device(s):"
			echo ${drives}
			{
				${PROG_MHDC} -v -q -w 0 -d ${drives} -c PGR3_TAKE ${ARG_FAILFAST}
				echo $? > /tmp/mhdc.$$.exit
			} | tee /tmp/mhdc.$$.out
			mhdc_exit_code="a"
			attempt=0
			while ! expr ${mhdc_exit_code} + 1 >/dev/null 2>&1;do
				attempt=`expr ${attempt} + 1`
				if [ ${attempt} -eq 10 ]; then
					break
				elif [ ${attempt} -gt 1 ]; then
					echo "Waiting for exit code in /tmp/mhdc.$$.exit"
					sleep 1
				fi
				mhdc_exit_code=`cat /tmp/mhdc.$$.exit`
			done
			if [ -f "/usr/bin/nawk" ]; then
				ok_drives="`nawk \"${awk_scr}\" /tmp/mhdc.$$.out`"
			else
				ok_drives="`awk \"${awk_scr}\" /tmp/mhdc.$$.out`"
			fi
			mhdc_cache_add ${service} pgr3="${drives}" success="${ok_drives}"
		else
			#
			# We have not found any PGR3 compliant disks
			dated_echo "PGR3: Device support test failed!" ERROR
			mhdc_exit_code=10
		fi
	else
		# SCSI-2 reservations
		dated_echo "Placing SCSI-2 reserve on device(s) ${drives}"
		if [ "${PROP_SCSI2_THREADED_RESERVE}" = "${BOOL_STR_TRUE}" -a "${PROP_SCSI2_THREADED_RESERVE_COUNT}" != "0" ]; then
			if [ $drive_cnt -gt ${PROP_SCSI2_THREADED_RESERVE_COUNT} -a ${PROP_SCSI2_THREADED_RESERVE_COUNT} -gt 0 ]; then
				drive_cnt=${PROP_SCSI2_THREADED_RESERVE_COUNT}
			fi
			SCSI2_THREADS="-j${drive_cnt}"
		else
			SCSI2_THREADS=
		fi
		{
			instance_limit ${MAX_CONCURRENT_TKOWNS} ${PROG_MHDC} -q -v -2 -w 0 ${SCSI2_THREADS} ${SCSI2_RETRIES} ${METRO_ARG} -d ${drives} -c TKOWN ${ARG_FAILFAST}
			echo $? > /tmp/mhdc.$$.exit
		} | tee /tmp/mhdc.$$.out
		mhdc_exit_code="a"
		attempt=0
		while ! expr ${mhdc_exit_code} + 1 >/dev/null 2>&1;do
			attempt=`expr ${attempt} + 1`
			if [ ${attempt} -eq 10 ]; then
				break
			elif [ ${attempt} -gt 1 ]; then
				echo "Waiting for exit code in /tmp/mhdc.$$.exit"
				sleep 1
			fi
			mhdc_exit_code=`cat /tmp/mhdc.$$.exit`
		done
		ok_drives="`nawk \"${awk_scr}\" /tmp/mhdc.$$.out`"
		mhdc_cache_add ${service} scsi2="${drives}" success="${ok_drives}"
	fi
	
	rm -f /tmp/mhdc.$$.exit /tmp/mhdc.$$.out
	
	drive_names="`echo ${drives} | sed -e 's#/dev/rdsk/##g' -e 's#s0##g'`"
	for drive in $drive_names; do
		if [ $mhdc_exit_code -eq 0 ]; then
			notify_watchers "${LOG_INFO}" "RSF_SERVICE_STARTUP" "service=${service} resource=RESERVATION disk=${drive} status=OK"
		else
			notify_watchers "${LOG_CRIT}" "RSF_SERVICE_STARTUP" "service=${service} resource=RESERVATION disk=${drive} status=FAILED"
		fi
	done
	
	return ${mhdc_exit_code}
}

release_drives()
{
	service=$1
	shift
	scheme=$1
	shift
	ignore_mhdc_return=$1
	shift
	
	drives=$@
	drive_cnt=$#
	
	if [ $drive_cnt -eq 0 ]; then
		dated_echo "Nothing to release!" ERROR
		return 1
	fi
	
	dated_echo "About to release drives: ${drives}"
	
	case ${scheme} in
		PGR-3)
			${PROG_MHDC} -v -q -d ${drives} -c PGR3_RELEASE ENFAILFAST 0
			mhdc_exit_code=$?
			mhdc_cache_remove ${service} pgr3="${drives}"
			;;
		SCSI-2)
			${PROG_MHDC} -v -q -2 ${SCSI2_RETRIES} -d ${drives} -c RELEASE ENFAILFAST 0
			mhdc_exit_code=$?
			mhdc_cache_remove ${service} scsi2="${drives}"
			;;
		*)
			dated_echo "release_drives(): Unknown reservation scheme: ${scheme}"
			return 1
			;;
	esac
	
	if [ "${ignore_mhdc_return}" = "${BOOL_STR_TRUE}" -a ${mhdc_exit_code} -ne 0 ]; then
		dated_echo "Ignoring MHDC exit code (${mhdc_exit_code})"
		mhdc_exit_code=0
	fi
	
	disc_names="`echo ${drives} | sed -e 's#/dev/rdsk/##g' -e 's#s0##g'`"
	for disc in $disc_names; do
		if [ ${mhdc_exit_code} -eq 0 ]; then
			notify_watchers "${LOG_INFO}" "RSF_SERVICE_SHUTDOWN" "service=${service} resource=RESERVATION disk=${disc} status=OK"
		else
			notify_watchers "${LOG_CRIT}" "RSF_SERVICE_SHUTDOWN" "service=${service} resource=RESERVATION disk=${disc} status=FAILED"
		fi
	done
	
	return ${mhdc_exit_code}
}

mhdc_cache_clear() {
	service=$1
	if [ -z "${service}" ]; then
		dated_echo "mhdc_cache_clear: No service name given" ERROR
		return 1
	fi
	dated_echo "Clearing MHDC cache for service ${service}"
	rotatelogs ${MHDC_CACHE} ${service}
	return 0
}

mhdc_cache_read() {
	if [ $# -ne 4 ]; then
		dated_echo "Usage: mhdc_cache_read <service> <pgr3_varname> <scsi2_varname> <reserved_varname>"
	fi

	service=$1
	pgr3_varname=$2
	s2_varname=$3
	reserved_varname=$4

	pgr3_discs=
	s2_discs=
	reserved_discs=
	if [ -f ${MHDC_CACHE}/${service} ]; then
		dated_echo "Reading reservation drive list from cache file (${MHDC_CACHE}/${service})"
		. ${MHDC_CACHE}/${service}
	else
		dated_echo "Cache file ${MHDC_CACHE}/${service} does not exist"
	fi
	eval ${pgr3_varname}=\"${pgr3_discs}\"
	eval ${s2_varname}=\"${s2_discs}\"
	eval ${reserved_varname}=\"${reserved_discs}\"

	test -f ${MHDC_CACHE}/${service}
	return $?
}

mhdc_cache_write() {
	service=$1
	pgr3=
	scsi2=
	success=
	shift
	while [ $# -gt 0 ]; do
		case "$1" in
			pgr3=*)
				pgr3="${pgr3} ${1#*=}"
				;;
			scsi2=*)
				scsi2="${scsi2} ${1#*=}"
				;;
			success=*)
				success="${success} ${1#*=}"
				;;
			*)
				dated_echo "mhdc_cache_write: Unknown argument (${1})"
				return 1
				;;
		esac
		shift
	done
	#
	# Remove duplicates
	#
	[ -n "${pgr3}" ] && pgr3="`echo ${pgr3}|xargs -n1|sort|uniq|xargs`"
	[ -n "${scsi2}" ] && scsi2="`echo ${scsi2}|xargs -n1|sort|uniq|xargs`"
	[ -n "${success}" ] && success="`echo ${success}|xargs -n1|sort|uniq|xargs`"
	#
	# Now create the cache file
	#
	[ -f ${MHDC_CACHE}/${service} ] && rotatelogs ${MHDC_CACHE} ${service}
	{
		echo "#"
		echo "# zpool ${service} device list as of `date`"
		if [ -n "${pgr3}" ]; then
			echo "pgr3_discs=\"${pgr3}\""
		elif [ -n "${scsi2}" ]; then
			echo "s2_discs=\"${scsi2}\""
		fi
		echo "reserved_discs=\"${success}\""
	} > ${MHDC_CACHE}/${service}

	return 0
}

mhdc_cache_add() {
	service=$1
	pgr3_to_add=
	scsi2_to_add=
	success_to_add=
	shift
	while [ $# -gt 0 ]; do
		case "$1" in
			pgr3=*)
				pgr3_to_add="${pgr3_to_add} ${1#*=}"
				;;
			scsi2=*)
				scsi2_to_add="${scsi2_to_add} ${1#*=}"
				;;
			success=*)
				success_to_add="${success_to_add} ${1#*=}"
				;;
			*)
				dated_echo "mhdc_cache_add: Unknown argument (${1})"
				return 1
				;;
		esac
		shift
	done
	mhdc_cache_read ${service} pgr3 scsi2 success
	mhdc_cache_write ${service} \
		pgr3="${pgr3} ${pgr3_to_add}" \
		scsi2="${scsi2} ${scsi2_to_add}" \
		success="${success} ${success_to_add}"
}

mhdc_cache_remove() {
	service=$1
	pgr3_to_rem=
	scsi2_to_rem=
	shift
	while [ $# -gt 0 ]; do
		case "$1" in
			pgr3=*)
				pgr3_to_rem="${pgr3_to_rem} ${1#*=}"
				;;
			scsi2=*)
				scsi2_to_rem="${scsi2_to_rem} ${1#*=}"
				;;
			*)
				dated_echo "mhdc_cache_remove: Unknown argument (${1})"
				return 1
				;;
		esac
		shift
	done
	mhdc_cache_read ${service} pgr3 scsi2 success
	pgr3_final=
	scsi2_final=
	success_final=
	changed=no
	for disk in ${pgr3}; do
		if expr " ${pgr3_to_rem} " : ".* ${disk} .*" 2>/dev/null; then
			changed=yes
			continue
		fi
		pgr3_final="${pgr3_final} ${disk}"
	done
	for disk in ${scsi2}; do
		if expr " ${scsi2_to_rem} " : ".* ${disk} .*" 2>/dev/null; then
			changed=yes
			continue
		fi
		scsi2_final="${scsi2_final} ${disk}"
	done
	for disk in ${success}; do
		if expr " ${pgr3_to_rem} ${scsi2_to_rem} " : ".* ${disk} .*" 2>/dev/null; then
			changed=yes
			continue
		fi
		success_final="${success_final} ${disk}"
	done
	if [ "${changed}" = "yes" ]; then
		mhdc_cache_write ${service} \
			pgr3="${pgr3_final}" \
			scsi2="${scsi2_final}" \
			success="${success_final}"
	fi
}
